import json
import random
import time
from typing import Any, Dict, Generator, List, Optional, Union
from uuid import uuid4

from curl_cffi import requests

from webscout import exceptions
from webscout.AIbase import AISearch, SearchResponse
from webscout.litagent import LitAgent
from webscout.sanitize import sanitize_stream


class Perplexity(AISearch):
    """A class to interact with the Perplexity AI search API.

    Perplexity provides a powerful search interface that returns AI-generated responses
    based on web content. It supports both streaming and non-streaming responses,
    multiple search modes, and model selection.

    Basic Usage:
        >>> from webscout import Perplexity
        >>> ai = Perplexity()
        >>> # Non-streaming example
        >>> response = ai.search("What is Python?")
        >>> print(response)
        Python is a high-level programming language...

        >>> # Streaming example
        >>> for chunk in ai.search("Tell me about AI", stream=True):
        ...     print(chunk, end="", flush=True)
        Artificial Intelligence is...
    """

    def __init__(
        self,
        cookies: Optional[Dict[str, str]] = None,
        timeout: int = 60,
        proxies: Optional[Dict[str, str]] = None,
    ):
        """
        Initialize the Perplexity client.
        """
        self.timeout = timeout
        self.agent = LitAgent()
        self.headers = {
            "accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7",
            "accept-language": "en-US,en;q=0.9",
            "cache-control": "max-age=0",
            "dnt": "1",
            "priority": "u=0, i",
            "sec-ch-ua": '"Not;A=Brand";v="24", "Chromium";v="128"',
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": '"Windows"',
            "sec-fetch-dest": "document",
            "sec-fetch-mode": "navigate",
            "sec-fetch-site": "same-origin",
            "sec-fetch-user": "?1",
            "upgrade-insecure-requests": "1",
            "user-agent": self.agent.random(),
        }
        self.session = requests.Session(
            headers=self.headers, cookies=cookies or {}, impersonate="chrome"
        )

        if proxies:
            from typing import cast
            self.session.proxies.update(cast(Any, proxies))

        # Initialize session info if possible
        try:
            self.timestamp = format(random.getrandbits(32), "08x")
            response = self.session.get(
                f"https://www.perplexity.ai/socket.io/?EIO=4&transport=polling&t={self.timestamp}"
            )
            if response.status_code == 200 and response.text.startswith("0"):
                self.sid = json.loads(response.text[1:])["sid"]
                self.session.post(
                    f"https://www.perplexity.ai/socket.io/?EIO=4&transport=polling&t={self.timestamp}&sid={self.sid}",
                    data='40{"jwt":"anonymous-ask-user"}',
                )
            self.session.get("https://www.perplexity.ai/api/auth/session")
        except Exception:
            pass

    def _extract_answer(self, response):
        """
        Extract the answer from the response.
        """
        if not response:
            return ""

        # Handle FINAL step at top level
        if response.get("step_type") == "FINAL" and "content" in response:
            content = response["content"]
            if isinstance(content, dict) and "answer" in content:
                answer_content = content["answer"]
                if isinstance(answer_content, str):
                    try:
                        return json.loads(answer_content).get("answer", "")
                    except Exception:
                        return str(answer_content)
                elif isinstance(answer_content, dict):
                    return answer_content.get("answer", "")

        # New pattern extraction
        if "text" in response:
            text_val = response["text"]
            if isinstance(text_val, list):
                for step in text_val:
                    if step.get("type") == "answer" and "value" in step:
                        return step["value"]
                    if step.get("step_type") == "FINAL" and "content" in step:
                        content = step["content"]
                        if isinstance(content, dict) and "answer" in content:
                            answer_content = content["answer"]
                            if isinstance(answer_content, str):
                                try:
                                    return json.loads(answer_content).get("answer", "")
                                except Exception:
                                    return str(answer_content)
                            elif isinstance(answer_content, dict):
                                return answer_content.get("answer", "")
            elif isinstance(text_val, str):
                try:
                    return json.loads(text_val).get("answer", text_val)
                except Exception:
                    return text_val
        return ""

        # New pattern extraction
        if "text" in response:
            text_val = response["text"]
            if isinstance(text_val, list):
                for step in text_val:
                    if step.get("type") == "answer" and "value" in step:
                        return step["value"]
                    if step.get("step_type") == "FINAL" and "content" in step:
                        content = step["content"]
                        if isinstance(content, dict) and "answer" in content:
                            answer_content = content["answer"]
                            if isinstance(answer_content, str):
                                try:
                                    return json.loads(answer_content).get("answer", "")
                                except Exception:
                                    return str(answer_content)
                            elif isinstance(answer_content, dict):
                                return answer_content.get("answer", "")
            elif isinstance(text_val, str):
                try:
                    return json.loads(text_val).get("answer", text_val)
                except Exception:
                    return text_val
        return ""

        # New pattern extraction
        if "text" in response:
            text_val = response["text"]
            if isinstance(text_val, list):
                for step in text_val:
                    if step.get("type") == "answer" and "value" in step:
                        return step["value"]
                    if step.get("step_type") == "FINAL" and "content" in step:
                        content = step["content"]
                        if isinstance(content, dict) and "answer" in content:
                            try:
                                return json.loads(content["answer"]).get("answer", "")
                            except Exception:
                                return str(content["answer"])
            elif isinstance(text_val, str):
                try:
                    return json.loads(text_val).get("answer", text_val)
                except Exception:
                    return text_val
        return ""

    def search(
        self,
        prompt: str,
        stream: bool = False,
        raw: bool = False,
        **kwargs: Any,
    ) -> Union[SearchResponse, Generator[Union[Dict[str, str], SearchResponse], None, None], List[Any], Dict[str, Any], str]:
        """Search using the Perplexity API and get AI-generated responses."""
        sources = kwargs.get("sources", ["web"])
        follow_up = kwargs.get("follow_up")
        incognito = kwargs.get("incognito", True)
        language = kwargs.get("language", "en-US")
        mode = kwargs.get("mode", "auto")
        model = kwargs.get("model")
        max_retries = kwargs.get("max_retries", 3)

        # Prepare request data
        json_data: Dict[str, Any] = {
            "query_str": prompt,
            "params": {
                "attachments": follow_up["attachments"] if follow_up else [],
                "frontend_context_uuid": str(uuid4()),
                "frontend_uuid": str(uuid4()),
                "is_incognito": incognito,
                "language": language,
                "last_backend_uuid": follow_up["backend_uuid"] if follow_up else None,
                "mode": "concise" if mode == "auto" else "copilot",
                "model_preference": "turbo",  # Default
                "source": "default",
                "sources": sources,
                "version": "2.18",
            },
        }

        # Handle model selection if provided
        if model:
            json_data["params"]["model_preference"] = model

        for attempt in range(max_retries):
            try:
                resp = self.session.post(
                    "https://www.perplexity.ai/rest/sse/perplexity_ask",
                    json=json_data,
                    stream=True,
                    timeout=self.timeout,
                )

                if resp.status_code == 502:
                    if attempt < max_retries - 1:
                        time.sleep(2**attempt)
                        continue
                    raise exceptions.APIConnectionError(
                        f"Perplexity API returned 502 Bad Gateway after {max_retries} attempts."
                    )

                if resp.status_code != 200:
                    raise exceptions.APIConnectionError(
                        f"API returned status code {resp.status_code}: {resp.text}"
                    )

                def stream_response():
                    full_text = ""
                    def extract_perplexity_content(data):
                        nonlocal full_text
                        current_answer = self._extract_answer(data)
                        if current_answer and len(current_answer) > len(full_text):
                            delta = current_answer[len(full_text):]
                            full_text = current_answer
                            return delta
                        return None

                    processed_chunks = sanitize_stream(
                        data=resp.iter_content(chunk_size=1024),
                        to_json=True,
                        extract_regexes=[r"data:\s*({.*})"],
                        content_extractor=lambda chunk: extract_perplexity_content(chunk),
                        yield_raw_on_error=False,
                        encoding='utf-8',
                        encoding_errors='replace',
                        line_delimiter="\r\n\r\n",
                        raw=raw,
                        output_formatter=None if raw else lambda x: SearchResponse(x) if isinstance(x, str) else x,
                    )

                    yield from processed_chunks

                if stream:
                    return stream_response()
                else:
                    if raw:
                        return resp.text

                    full_response_text = ""
                    for chunk in stream_response():
                        full_response_text += str(chunk)

                    return SearchResponse(full_response_text)

            except requests.RequestsError as e:
                if attempt < max_retries - 1:
                    time.sleep(2**attempt)
                    continue
                raise exceptions.APIConnectionError(f"Connection error: {str(e)}")
            except Exception as e:
                raise exceptions.FailedToGenerateResponseError(f"Error: {str(e)}")

        raise exceptions.FailedToGenerateResponseError(
            "Failed to get response after multiple attempts"
        )


if __name__ == "__main__":
    ai = Perplexity()
    response = ai.search("What is Python?", stream=False)
    print(response)



import json
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Callable, Dict, Generator, List, Optional, TypedDict, Union, cast

import requests
from litprinter import ic

from webscout.AIbase import ModelList, Response, SimpleModelList
from webscout.model_fetcher import BackgroundModelFetcher

# Import the utils for response structures
from webscout.Provider.Openai_comp.utils import ChatCompletion, ChatCompletionChunk


# Define tool-related structures
class ToolDefinition(TypedDict):
    """Definition of a tool that can be called by the AI"""
    type: str
    function: Dict[str, Any]

class FunctionParameters(TypedDict):
    """Parameters for a function"""
    type: str
    properties: Dict[str, Dict[str, Any]]
    required: List[str]

class FunctionDefinition(TypedDict):
    """Definition of a function that can be called by the AI"""
    name: str
    description: str
    parameters: FunctionParameters

@dataclass
class Tool:
    """Tool class that can be passed to the provider"""
    name: str
    description: str
    parameters: Dict[str, Dict[str, Any]]
    required_params: Optional[List[str]] = None
    implementation: Optional[Callable] = None

    def to_dict(self) -> ToolDefinition:
        """Convert to OpenAI tool definition format"""
        function_def = {
            "name": self.name,
            "description": self.description,
            "parameters": {
                "type": "object",
                "properties": self.parameters,
                "required": self.required_params or list(self.parameters.keys())
            }
        }

        return {
            "type": "function",
            "function": function_def
        }

    def execute(self, arguments: Dict[str, Any]) -> Any:
        """Execute the tool with the given arguments"""
        if not self.implementation:
            return f"Tool '{self.name}' does not have an implementation."

        try:
            return self.implementation(**arguments)
        except Exception as e:
            ic.configureOutput(prefix='ERROR| ')
            ic(f"Error executing tool '{self.name}': {str(e)}")
            return f"Error executing tool '{self.name}': {str(e)}"

class BaseCompletions(ABC):
    @abstractmethod
    def create(
        self,
        *,
        model: str,
        messages: List[Dict[str, Any]],  # Changed to Any to support complex message structures
        max_tokens: Optional[int] = None,
        stream: bool = False,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        tools: Optional[List[Union[Tool, Dict[str, Any]]]] = None,  # Support for tool definitions
        tool_choice: Optional[Union[str, Dict[str, Any]]] = None,  # Support for tool_choice parameter
        timeout: Optional[int] = None,
        proxies: Optional[dict] = None,
        **kwargs: Any
    ) -> Union[ChatCompletion, Generator[ChatCompletionChunk, None, None]]:
        """
        Abstract method to create chat completions with tool support.

        Args:
            model: The model to use for completion
            messages: List of message dictionaries
            max_tokens: Maximum number of tokens to generate
            stream: Whether to stream the response
            temperature: Sampling temperature
            top_p: Nucleus sampling parameter
            tools: List of tool definitions available for the model to use
            tool_choice: Control over which tool the model should use
            **kwargs: Additional model-specific parameters

        Returns:
            Either a completion object or a generator of completion chunks if streaming
        """
        raise NotImplementedError

    def format_tool_calls(self, tools: List[Union[Tool, Dict[str, Any]]]) -> List[Dict[str, Any]]:
        """Convert tools to the format expected by the provider"""
        formatted_tools = []

        for tool in tools:
            if isinstance(tool, Tool):
                formatted_tools.append(tool.to_dict())
            elif isinstance(tool, dict):
                # Assume already formatted correctly
                formatted_tools.append(tool)
            else:
                ic.configureOutput(prefix='WARNING| ')
                ic(f"Skipping invalid tool type: {type(tool)}")

        return formatted_tools

    def process_tool_calls(self, tool_calls: List[Dict[str, Any]], available_tools: Dict[str, Tool]) -> List[Dict[str, Any]]:
        """
        Process tool calls and execute the relevant tools.

        Args:
            tool_calls: List of tool calls from the model
            available_tools: Dictionary mapping tool names to their implementations

        Returns:
            List of results from executing the tools
        """
        results = []

        for call in tool_calls:
            try:
                function_call = call.get("function", {})
                tool_name = function_call.get("name")
                arguments_str = function_call.get("arguments", "{}")

                # Parse arguments
                try:
                    if isinstance(arguments_str, str):
                        arguments = json.loads(arguments_str)
                    else:
                        arguments = arguments_str
                except json.JSONDecodeError:
                    results.append({
                        "tool_call_id": call.get("id"),
                        "result": f"Error: Could not parse arguments JSON: {arguments_str}"
                    })
                    continue

                # Execute the tool if available
                if tool_name in available_tools:
                    tool_result = available_tools[tool_name].execute(arguments)
                    results.append({
                        "tool_call_id": call.get("id"),
                        "result": str(tool_result)
                    })
                else:
                    results.append({
                        "tool_call_id": call.get("id"),
                        "result": f"Error: Tool '{tool_name}' not found."
                    })
            except Exception as e:
                ic.configureOutput(prefix='ERROR| ')
                ic(f"Error processing tool call: {str(e)}")
                results.append({
                    "tool_call_id": call.get("id", "unknown"),
                    "result": f"Error processing tool call: {str(e)}"
                })

        return results


class BaseChat(ABC):
    completions: BaseCompletions


class OpenAICompatibleProvider(ABC):
    """
    Abstract Base Class for providers mimicking the OpenAI Python client structure.
    Requires a nested 'chat.completions' structure with tool support.
    Users can provide their own proxies via the proxies parameter.
    """
    chat: BaseChat
    available_tools: Dict[str, Tool] = {}  # Dictionary of available tools
    supports_tools: bool = False  # Whether the provider supports tools
    supports_tool_choice: bool = False  # Whether the provider supports tool_choice
    required_auth: bool = False  # Whether the provider requires authentication

    def __init__(self, api_key: Optional[str] = None, tools: Optional[List[Tool]] = None, proxies: Optional[dict] = None, **kwargs: Any):
        self.available_tools = {}
        if tools:
            self.register_tools(tools)
        self.proxies = proxies or {}
        self.session = requests.Session()
        if self.proxies:
            self.session.proxies.update(self.proxies)
        # raise NotImplementedError  # <-- Commented out for metaclass test

    @property
    @abstractmethod
    def models(self) -> ModelList:
        """
        Property that returns an object with a .list() method returning available models.
        Subclasses must implement this property.
        """
        raise NotImplementedError

    def register_tools(self, tools: List[Tool]) -> None:
        """
        Register tools with the provider.

        Args:
            tools: List of Tool objects to register
        """
        for tool in tools:
            self.available_tools[tool.name] = tool

    def _start_background_model_fetch(self, api_key: Optional[str] = None) -> None:
        """Start non-blocking background fetch of available models.

        Fetches models from the provider's API in a background thread and caches
        the result. Immediately returns and allows initialization to continue.

        Args:
            api_key: Optional API key for authentication during model fetch.
        """
        # Use class name as provider identifier for caching
        provider_name = self.__class__.__name__

        # Define the fetch function that will run in the background
        def fetch_models() -> List[str]:
            """Fetch models via the classmethod and update cache/class var."""
            try:
                get_models_func = cast(
                    Callable[[Optional[str]], List[str]],
                    getattr(self.__class__, "get_models", None)
                )
                if get_models_func is not None:
                    models = get_models_func(api_key)
                    if models and isinstance(models, list):
                        # Also call update_available_models if available to update
                        # the class variable
                        update_func = cast(
                            Callable[[Optional[str]], None],
                            getattr(self.__class__, "update_available_models", None)
                        )
                        if update_func is not None:
                            update_func(api_key)
                        return models
            except Exception as e:
                ic(f"Failed to fetch models for {provider_name}: {e}")
            return []

        # Get current available models as fallback
        fallback_models = getattr(self.__class__, "AVAILABLE_MODELS", [])

        # Start async fetch
        fetcher = BackgroundModelFetcher()
        fetcher.fetch_async(
            provider_name=provider_name,
            fetch_func=fetch_models,
            fallback_models=list(fallback_models) if fallback_models else [],
        )

    def get_tool_by_name(self, name: str) -> Optional[Tool]:
        """Get a tool by name"""
        return self.available_tools.get(name)

    def format_tool_response(self, messages: List[Dict[str, Any]], tool_results: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """
        Format tool results as messages to be sent back to the provider.

        Args:
            messages: The original messages
            tool_results: Results from executing tools

        Returns:
            Updated message list with tool results
        """
        updated_messages = messages.copy()

        # Find the assistant message with tool calls
        for i, msg in enumerate(reversed(updated_messages)):
            if msg.get("role") == "assistant" and "tool_calls" in msg:
                # For each tool result, add a tool message
                for result in tool_results:
                    tool_message = {
                        "role": "tool",
                        "tool_call_id": result["tool_call_id"],
                        "content": result["result"]
                    }
                    updated_messages.append(tool_message)
                break

        return updated_messages



"""Base plugin system for SwiftCLI."""

from typing import Any, Dict, List, Optional


class Plugin:
    """
    Base class for SwiftCLI plugins.

    Plugins can extend the CLI functionality by hooking into various stages
    of command execution. Subclass this class and override the methods
    you want to hook into.

    Attributes:
        app: The CLI application instance
        enabled: Whether the plugin is enabled
        config: Plugin configuration dictionary

    Example:
        class LoggingPlugin(Plugin):
            def before_command(self, command, args):
                print(f"[LOG] Running command: {command}")

            def after_command(self, command, args, result):
                print(f"[LOG] Command completed: {result}")
    """

    def __init__(self):
        self.app = None  # Set by plugin manager
        self.enabled: bool = True
        self.config: Dict[str, Any] = {}

    def init_app(self, app: Any) -> None:
        """
        Initialize plugin with CLI app instance.

        Args:
            app: The CLI application instance
        """
        self.app = app

    def before_command(self, command: str, args: List[str]) -> Optional[bool]:
        """
        Called before command execution.

        Args:
            command: Command name
            args: Command arguments

        Returns:
            Optional[bool]: Return False to prevent command execution
        """
        pass

    def after_command(self, command: str, args: List[str], result: Any) -> None:
        """
        Called after command execution.

        Args:
            command: Command name
            args: Command arguments
            result: Command result
        """
        pass

    def on_error(self, command: str, error: Exception) -> None:
        """
        Called when command raises an error.

        Args:
            command: Command name
            error: The exception that was raised
        """
        pass

    def on_help(self, command: str) -> Optional[str]:
        """
        Called when help is requested for a command.

        Args:
            command: Command name

        Returns:
            Optional[str]: Additional help text to display
        """
        pass

    def on_completion(self, command: str, incomplete: str) -> List[str]:
        """
        Called when shell completion is requested.

        Args:
            command: Command name
            incomplete: Incomplete text to complete

        Returns:
            List[str]: Possible completions
        """
        return []

    def configure(self, config: Dict[str, Any]) -> None:
        """
        Configure the plugin.

        Args:
            config: Configuration dictionary
        """
        self.config.update(config)

    def validate_config(self) -> None:
        """
        Validate plugin configuration.

        Raises:
            PluginError: If configuration is invalid
        """
        pass

    def enable(self) -> None:
        """Enable the plugin."""
        self.enabled = True

    def disable(self) -> None:
        """Disable the plugin."""
        self.enabled = False

    @property
    def name(self) -> str:
        """Get plugin name."""
        return self.__class__.__name__

    def __repr__(self) -> str:
        status = "enabled" if self.enabled else "disabled"
        return f"<{self.name} [{status}]>"



import json
from typing import Any, Dict, Generator, Optional, Union, cast

from curl_cffi import CurlError
from curl_cffi.requests import Session

from webscout import exceptions
from webscout.AIbase import Provider, Response
from webscout.AIutel import AwesomePrompts, Conversation, Optimizers, sanitize_stream


class ChatSandbox(Provider):
    """
    Sends a chat message to the specified model via the chatsandbox API.

    This provider allows you to interact with various AI models through the chatsandbox.com
    interface, supporting different models/models like OpenAI, DeepSeek, Llama, etc.

    Attributes:
        model (str): The model to chat with (e.g., "openai", "deepseek", "llama").

    Examples:
        >>> from webscout.Provider.ChatSandbox import ChatSandbox
        >>> ai = ChatSandbox(model="openai")
        >>> response = ai.chat("Hello, how are you?")
        >>> print(response)
        'I'm doing well, thank you for asking! How can I assist you today?'
    """
    required_auth = False
    AVAILABLE_MODELS = [
        "openai",
        "openai-gpt-4o",
        "openai-o1-mini",
        "deepseek",
        "deepseek-r1",
        "deepseek-r1-full",
        "gemini",
        "gemini-thinking",
        "mistral",
        "mistral-large",
        "gemma-3",
        "llama"
    ]


    def __init__(
        self,
        model: str = "openai",
        is_conversation: bool = True,
        max_tokens: int = 600,
        timeout: int = 30,
        intro: Optional[str] = None,
        filepath: Optional[str] = None,
        update_file: bool = True,
        proxies: dict = {},
        history_offset: int = 10250,
        act: Optional[str] = None,
    ):
        """
        Initializes the ChatSandbox API with given parameters.

        Args:
            model (str): The model to chat with (e.g., "openai", "deepseek", "llama").
            is_conversation (bool): Whether the provider is in conversation mode.
            max_tokens (int): Maximum number of tokens to sample.
            timeout (int): Timeout for API requests.
            intro (str): Introduction message for the conversation.
            filepath (str): Filepath for storing conversation history.
            update_file (bool): Whether to update the conversation history file.
            proxies (dict): Proxies for the API requests.
            history_offset (int): Offset for conversation history.
            act (str): Act for the conversation.
        """
        if model not in self.AVAILABLE_MODELS:
            raise ValueError(f"Invalid model: {model}. Choose from: {self.AVAILABLE_MODELS}")

        # Initialize curl_cffi Session with impersonation
        self.session = Session(impersonate="chrome120")
        self.model = model
        self.is_conversation = is_conversation
        self.max_tokens_to_sample = max_tokens
        self.api_endpoint = "https://chatsandbox.com/api/chat"
        self.timeout = timeout
        self.last_response = {}

        # Set up headers
        self.headers = {
            'accept': '*/*',
            'accept-language': 'en-US,en;q=0.9',
            'content-type': 'application/json',
            'origin': 'https://chatsandbox.com',
            'referer': f'https://chatsandbox.com/chat/{self.model}',
        }

        # Update curl_cffi session headers and proxies
        self.session.headers.update(self.headers)
        if proxies:
            self.session.proxies.update(proxies)

        self.__available_optimizers = (
            method
            for method in dir(Optimizers)
            if callable(getattr(Optimizers, method)) and not method.startswith("__")
        )

        self.conversation = Conversation(
            is_conversation, self.max_tokens_to_sample, filepath, update_file
        )
        self.conversation.history_offset = history_offset

        if act:
            self.conversation.intro = AwesomePrompts().get_act(cast(Union[str, int], act), default=self.conversation.intro, case_insensitive=True
            ) or self.conversation.intro
        elif intro:
            self.conversation.intro = intro

    @staticmethod
    def _chatsandbox_extractor(chunk: Union[str, Dict[str, Any]]) -> Optional[str]:
        """Extracts content from the chatsandbox stream format."""
        if isinstance(chunk, str):
            try:
                data = json.loads(chunk)
                if isinstance(data, dict):
                    if "reasoning_content" in data and data["reasoning_content"]:
                        return data["reasoning_content"]
                    if "content" in data and data["content"]:
                        return data["content"]
                return chunk
            except json.JSONDecodeError:
                return chunk
        return None

    def ask(
        self,
        prompt: str,
        stream: bool = False,
        raw: bool = False,
        optimizer: Optional[str] = None,
        conversationally: bool = False,
        **kwargs: Any,
    ) -> Response:
        """
        Sends a prompt to the ChatSandbox API and returns the response.

        Args:
            prompt (str): The prompt to send to the API.
            stream (bool): Whether to stream the response.
            raw (bool): Whether to return the raw response.
            optimizer (str): Optimizer to use for the prompt.
            conversationally (bool): Whether to generate the prompt conversationally.

        Returns:
            Union[Dict[str, Any], Generator]: The API response.
        """
        conversation_prompt = self.conversation.gen_complete_prompt(prompt)
        if optimizer:
            if optimizer in self.__available_optimizers:
                conversation_prompt = getattr(Optimizers, optimizer)(
                    conversation_prompt if conversationally else prompt
                )
            else:
                raise Exception(
                    f"Optimizer is not one of {self.__available_optimizers}"
                )

        # Prepare the payload
        payload = {
            "messages": [conversation_prompt],
            "character": self.model
        }

        def for_stream():
            try:
                response = self.session.post(
                    self.api_endpoint,
                    json=payload,
                    stream=True,
                    timeout=self.timeout
                )
                if not response.ok:
                    raise exceptions.FailedToGenerateResponseError(
                        f"Failed to generate response - ({response.status_code}, {response.reason}) - {response.text}"
                    )

                streaming_response = ""
                # Use sanitize_stream with the custom extractor
                processed_stream = sanitize_stream(
                    data=response.iter_content(chunk_size=None),  # Pass byte iterator
                    intro_value=None,
                    to_json=False,
                    content_extractor=self._chatsandbox_extractor,
                    raw=raw,
                )

                for content_chunk in processed_stream:
                    if raw:
                        yield content_chunk
                    else:
                        if content_chunk and isinstance(content_chunk, str):
                            streaming_response += content_chunk
                            yield dict(text=content_chunk)

                self.last_response.update(dict(text=streaming_response))
                self.conversation.update_chat_history(
                    prompt, self.get_message(self.last_response)
                )
            except CurlError as e:
                raise exceptions.FailedToGenerateResponseError(f"Request failed (CurlError): {e}")
            except Exception as e:
                raise exceptions.FailedToGenerateResponseError(f"An unexpected error occurred ({type(e).__name__}): {e}")

        def for_non_stream():
            for _ in for_stream():
                pass
            return self.last_response

        return for_stream() if stream else for_non_stream()

    def chat(
        self,
        prompt: str,
        stream: bool = False,
        optimizer: Optional[str] = None,
        conversationally: bool = False,
        **kwargs: Any,
    ) -> Union[str, Generator[str, None, None]]:
        """
        Generates a response from the ChatSandbox API.

        Args:
            prompt (str): The prompt to send to the API.
            stream (bool): Whether to stream the response.
            optimizer (str): Optimizer to use for the prompt.
            conversationally (bool): Whether to generate the prompt conversationally.
            **kwargs: Additional parameters including raw.

        Returns:
            str: The API response.
        """
        raw = kwargs.get("raw", False)
        def for_stream():
            for response in self.ask(
                prompt,
                stream=True,
                raw=raw,
                optimizer=optimizer,
                conversationally=conversationally,
            ):
                if raw:
                    yield response
                else:
                    yield self.get_message(response)

        if stream:
            return for_stream()
        else:
            result = self.ask(
                prompt,
                stream=False,
                raw=raw,
                optimizer=optimizer,
                conversationally=conversationally,
            )
            if raw:
                return cast(str, result)
            else:
                return self.get_message(result)

    def get_message(self, response: Response) -> str:
        """
        Extract the message from the API response.

        Args:
            response (Response): The API response.

        Returns:
            str: The extracted message.
        """
        if not isinstance(response, dict):
            return str(response)

        raw_text = cast(Dict[str, Any], response).get("text", "")

        # Try to parse as JSON
        try:
            data = json.loads(raw_text)
            if isinstance(data, dict):
                if "reasoning_content" in data and data["reasoning_content"]:
                    return data["reasoning_content"]
                elif "content" in data and data["content"]:
                    return data["content"]
                elif "message" in data:
                    return data["message"]
                elif "response" in data:
                    return data["response"]
                elif "text" in data:
                    return data["text"]
                return json.dumps(data, ensure_ascii=False)
        except json.JSONDecodeError:
            pass

        return raw_text.strip()

# --- Example Usage ---
if __name__ == "__main__":
    from rich import print
    print("-" * 80)
    print(f"{ 'Model':<50} {'Status':<10} {'Response'}")
    print("-" * 80)

    for model in ChatSandbox.AVAILABLE_MODELS:
        try:
            test_ai = ChatSandbox(model=model, timeout=60)
            response = test_ai.chat("Say 'Hello' in one word")

            if hasattr(response, "__iter__") and not isinstance(response, (str, bytes)):
                response_text = "".join(list(response))
            else:
                response_text = str(response)

            if response_text and len(response_text.strip()) > 0:
                status = "✓"
                display_text = response_text.strip()[:50].replace('\n', ' ') + ("..." if len(response_text.strip()) > 50 else "")
            else:
                status = "✗"
                display_text = "Empty or invalid response"
            print(f"{model:<50} {status:<10} {display_text}")
        except Exception as e:
            print(f"{model:<50} {'✗':<10} {str(e)}")



"""
Command-line interface for temporary email generation using SwiftCLI
"""

import sys
import time
from typing import Optional, cast

from rich.console import Console

from webscout.swiftcli import CLI, option

from .base import TempMailProvider, get_provider

# Initialize console for rich output
console = Console()

# Create CLI app
app = CLI(
    name="tempmail",
    help="Temporary Email CLI Tool",
    version="1.0.0"
)

@app.command()
@option("--wait", "-w", type=int, default=10, help="Wait time in seconds between checking for new messages")
@option("--output", "-o", type=str, choices=["simple", "detailed"], default="simple", help="Output format")
@option("--provider", "-p", type=str, choices=["mailtm", "tempmailio"], default="mailtm", help="Email provider to use")
def monitor(wait: int = 10, output: str = "simple", provider: str = "mailtm"):
    """
    Create and monitor a temporary email account

    Continuously checks for new messages and displays them in the terminal.
    Press Ctrl+C to exit and delete the account.
    """
    # Create a new email
    mail = cast(TempMailProvider, get_provider(provider_name=provider))

    if hasattr(mail, 'auto_create'):
        mail = cast(TempMailProvider, get_provider(provider_name=provider, async_provider=False))
        mail.create_account()
    else:
        success = mail.create_account()
        if not success:
            console.print("[red]Failed to create temporary email account.[/red]")
            sys.exit(1)

    console.print("[yellow]Temporary email address CLI[/yellow]")
    console.print(f"[cyan]EMAIL: {mail.email}[/cyan]")
    console.print(f"Account ID: {mail.account_id}")
    console.print("Press Ctrl+C to exit and delete the account")

    try:
        while True:
            console.print(f"Checking for new messages... ({time.strftime('%H:%M:%S')})")
            new_messages = mail.check_new_messages()

            if new_messages:
                console.print(f"New messages: {len(new_messages)}")
                console.print("-" * 40)

                for msg in new_messages:
                    console.print(f"From: {msg['from']}")
                    console.print(f"Subject: {msg['subject']}")

                    if output == "detailed":
                        console.print(f"Body: {msg['body']}")
                        console.print(f"Has attachments: {msg['hasAttachments']}")
                        console.print(f"Created at: {msg['createdAt']}")

                    console.print("[magenta]" + "-" * 40 + "[/magenta]")

            time.sleep(wait)

    except KeyboardInterrupt:
        console.print("\nCtrl+C pressed. Exiting program and deleting account...")
        if mail.delete_account():
            console.print("[red]--Account deleted--[/red]")
        else:
            console.print("[red]--Failed to delete account--[/red]")

@app.command()
@option("--provider", "-p", type=str, choices=["mailtm", "tempmailio"], default="mailtm", help="Email provider to use")
def create(provider: str = "mailtm"):
    """
    Create a temporary email account and display its details

    Creates a new temporary email account and displays the email address.
    The account will remain active until manually deleted.
    """
    mail = cast(TempMailProvider, get_provider(provider_name=provider))

    if hasattr(mail, 'auto_create'):
        mail = cast(TempMailProvider, get_provider(provider_name=provider, async_provider=False))
        mail.create_account()
    else:
        success = mail.create_account()
        if not success:
            console.print("[red]Failed to create temporary email account.[/red]")
            sys.exit(1)

    console.print("[yellow]Temporary Email Created[/yellow]")
    console.print(f"[cyan]EMAIL: {mail.email}[/cyan]")
    console.print(f"Account ID: {mail.account_id}")
    console.print("Use 'tempmail monitor' to check for messages")

@app.command()
@option("--email", "-e", type=str, required=True, help="Email address to check")
@option("--account-id", "-a", type=str, required=True, help="Account ID for the email")
@option("--provider", "-p", type=str, choices=["mailtm", "tempmailio"], default="mailtm", help="Email provider to use")
@option("--output", "-o", type=str, choices=["simple", "detailed"], default="simple", help="Output format")
def check(email: str, account_id: str, provider: str = "mailtm", output: str = "simple"):
    """
    Check messages for an existing temporary email account

    Retrieves and displays messages for an existing temporary email account.
    Requires the email address and account ID.
    """
    mail = cast(TempMailProvider, get_provider(provider_name=provider))
    mail.email = email
    mail.account_id = account_id

    # Try to get token if needed
    if hasattr(mail, 'get_token'):
        mail.get_token()

    console.print(f"[cyan]Checking messages for: {email}[/cyan]")

    try:
        messages = mail.get_messages()

        if not messages:
            console.print("[yellow]No messages found[/yellow]")
            return

        console.print(f"Found {len(messages)} messages:")
        console.print("-" * 40)

        for msg in messages:
            console.print(f"From: {msg['from']}")
            console.print(f"Subject: {msg['subject']}")

            if output == "detailed":
                console.print(f"Body: {msg['body']}")
                console.print(f"Has attachments: {msg['hasAttachments']}")
                console.print(f"Created at: {msg['createdAt']}")

            console.print("[magenta]" + "-" * 40 + "[/magenta]")

    except Exception as e:
        console.print(f"[red]Error checking messages: {str(e)}[/red]")

@app.command()
@option("--email", "-e", type=str, required=True, help="Email address to delete")
@option("--account-id", "-a", type=str, required=True, help="Account ID for the email")
@option("--provider", "-p", type=str, choices=["mailtm", "tempmailio"], default="mailtm", help="Email provider to use")
def delete(email: str, account_id: str, provider: str = "mailtm"):
    """
    Delete an existing temporary email account

    Deletes a temporary email account using the provided email address and account ID.
    """
    mail = cast(TempMailProvider, get_provider(provider_name=provider))
    mail.email = email
    mail.account_id = account_id

    # Try to get token if needed
    if hasattr(mail, 'get_token'):
        mail.get_token()

    console.print(f"[yellow]Deleting account: {email}[/yellow]")

    try:
        success = mail.delete_account()
        if success:
            console.print("[green]Account successfully deleted[/green]")
        else:
            console.print("[red]Failed to delete account[/red]")
    except Exception as e:
        console.print(f"[red]Error deleting account: {str(e)}[/red]")

def main():
    """Main CLI function for tempmail"""
    try:
        app.run()
    except Exception as e:
        console.print(f"[red]Error: {str(e)}[/red]")
        sys.exit(1)

if __name__ == "__main__":
    main()



import uuid
from typing import Any, Dict, Generator, Optional, Union, cast

from curl_cffi import CurlError
from curl_cffi.requests import Session

from webscout import exceptions
from webscout.AIbase import Provider, Response
from webscout.AIutel import AwesomePrompts, Conversation, Optimizers, sanitize_stream
from webscout.litagent import LitAgent


class QodoAI(Provider):
    """
    A class to interact with the Qodo AI API.
    """

    AVAILABLE_MODELS = [
        "gpt-4.1",
        "gpt-4o",
        "o3",
        "o4-mini",
        "claude-4-sonnet",
        "gemini-2.5-pro",
        "grok-4"

    ]

    @staticmethod
    def _qodo_extractor(chunk: Union[str, Dict[str, Any]]) -> Optional[str]:
        """Extracts content from Qodo stream JSON objects."""
        if isinstance(chunk, dict):
            data = chunk.get("data", {})
            if isinstance(data, dict):
                tool_args = data.get("tool_args", {})
                if isinstance(tool_args, dict):
                    return tool_args.get("content")
        return None

    def __init__(
        self,
        api_key: Optional[str] = None,
        is_conversation: bool = True,
        max_tokens: int = 2049,
        timeout: int = 30,
        intro: Optional[str] = None,
        filepath: Optional[str] = None,
        update_file: bool = True,
        proxies: dict = {},
        history_offset: int = 10250,
        act: Optional[str] = None,
        model: str = "claude-4-sonnet",
        browser: str = "chrome"
    ):
        """Initializes the Qodo AI API client."""
        if model not in self.AVAILABLE_MODELS:
            raise ValueError(f"Invalid model: {model}. Choose from: {self.AVAILABLE_MODELS}")

        self.url = "https://api.cli.qodo.ai/v2/agentic/start-task"
        self.info_url = "https://api.cli.qodo.ai/v2/info/get-things"

        # Initialize LitAgent for user agent generation
        self.agent = LitAgent()
        self.fingerprint = self.agent.generate_fingerprint(browser)

        # Store API key
        self.api_key = api_key or "sk-dS7U-extxMWUxc8SbYYOuncqGUIE8-y2OY8oMCpu0eI-qnSUyH9CYWO_eAMpqwfMo7pXU3QNrclfZYMO0M6BJTM"

        # Generate session ID dynamically from API
        self.session_id = self._get_session_id()
        self.request_id = str(uuid.uuid4())

        # Use the fingerprint for headers
        self.headers = {
            "Accept": "text/plain",
            "Accept-Encoding": "gzip, deflate, br, zstd",
            "Accept-Language": self.fingerprint["accept_language"],
            "Authorization": f"Bearer {self.api_key}",
            "Connection": "close",
            "Content-Type": "application/json",
            "host": "api.cli.qodo.ai",
            "Request-id": self.request_id,
            "User-Agent": self.fingerprint["user_agent"],
        }

        # Initialize curl_cffi Session
        self.session = Session()
        # Add Session-id to headers after getting it from API
        self.headers["Session-id"] = self.session_id
        self.session.headers.update(self.headers)
        self.session.proxies.update(proxies)

        self.is_conversation = is_conversation
        self.max_tokens_to_sample = max_tokens
        self.timeout = timeout
        self.last_response = {}
        self.model = model

        self.__available_optimizers = (
            method
            for method in dir(Optimizers)
            if callable(getattr(Optimizers, method)) and not method.startswith("__")
        )
        self.conversation = Conversation(
            is_conversation, self.max_tokens_to_sample, filepath, update_file
        )
        act_prompt = (
            AwesomePrompts().get_act(cast(Union[str, int], act), default=None, case_insensitive=True
            )
            if act
            else intro
        )
        if act_prompt:
            self.conversation.intro = act_prompt
        self.conversation.history_offset = history_offset

    def refresh_identity(self, browser: Optional[str] = None):
        """
        Refreshes the browser identity fingerprint.

        Args:
            browser: Specific browser to use for the new fingerprint
        """
        browser = browser or self.fingerprint.get("browser_type", "chrome")
        self.fingerprint = self.agent.generate_fingerprint(browser)

        # Update headers with new fingerprint
        self.headers.update({
            "Accept-Language": self.fingerprint["accept_language"],
            "User-Agent": self.fingerprint["user_agent"],
        })

        # Update session headers
        for header, value in self.headers.items():
            self.session.headers[header] = value

        return self.fingerprint

    def _build_payload(self, prompt: str):
        """Build the payload for Qodo AI API."""
        return {
            "agent_type": "cli",
            "session_id": self.session_id,
            "user_data": {
                "extension_version": "0.7.2",
                "os_platform": "win32",
                "os_version": "v23.9.0",
                "editor_type": "cli"
            },
            "tools": {
                "web_search": [
                    {
                        "name": "web_search",
                        "description": "Searches the web and returns results based on the user's query (Powered by Nimble).",
                        "inputSchema": {
                            "type": "object",
                            "properties": {
                                "llm_description": {
                                    "default": "Searches the web and returns results based on the user's query.",
                                    "title": "Llm Description",
                                    "type": "string"
                                },
                                "query": {
                                    "description": "The search query to execute",
                                    "title": "Query",
                                    "type": "string"
                                }
                            },
                            "required": ["query"],
                            "title": "NimbleWebSearch"
                        },
                        "be_tool": True,
                        "autoApproved": True
                    },
                    {
                        "name": "web_fetch",
                        "description": "Fetches content from a given URL (Powered by Nimble).",
                        "inputSchema": {
                            "type": "object",
                            "properties": {
                                "llm_description": {
                                    "default": "Fetches content from a given URL.",
                                    "title": "Llm Description",
                                    "type": "string"
                                },
                                "url": {
                                    "description": "The URL to fetch content from",
                                    "title": "Url",
                                    "type": "string"
                                }
                            },
                            "required": ["url"],
                            "title": "NimbleWebFetch"
                        },
                        "be_tool": True,
                        "autoApproved": True
                    }
                ]
            },
            # "projects_root_path": ["C:\\Users\\koula"],
            # "cwd": "C:\\Users\\koula",
            "user_request": prompt,
            "execution_strategy": "act",
            "custom_model": self.model,
            "stream": True
        }

    def ask(
        self,
        prompt: str,
        stream: bool = False,
        raw: bool = False,
        optimizer: Optional[str] = None,
        conversationally: bool = False,
        **kwargs: Any,
    ) -> Union[Dict[str, Any], Generator]:
        conversation_prompt = self.conversation.gen_complete_prompt(prompt)
        if optimizer:
            if optimizer in self.__available_optimizers:
                conversation_prompt = getattr(Optimizers, optimizer)(
                    conversation_prompt if conversationally else prompt
                )
            else:
                raise Exception(f"Optimizer is not one of {self.__available_optimizers}")

        payload = self._build_payload(conversation_prompt)
        payload["stream"] = stream


        def for_stream():
            try:
                response = self.session.post(
                    self.url,
                    json=payload,
                    stream=True,
                    timeout=self.timeout,
                    impersonate=self.fingerprint.get("browser_type", "chrome110")
                )
                # Check for internal server error with session ID in the response
                if response.status_code == 500 and response.text and "Internal server error, session ID:" in response.text:
                    # Switch to continue-task endpoint and retry
                    self.url = "https://api.cli.qodo.ai/v2/agentic/continue-task"
                    response = self.session.post(
                        self.url,
                        json=payload,
                        stream=True,
                        timeout=self.timeout,
                        impersonate=self.fingerprint.get("browser_type", "chrome110")
                    )
                if response.status_code == 401:
                    raise exceptions.FailedToGenerateResponseError(
                        "Invalid API key. You need to provide your own API key.\n"
                        "Usage: QodoAI(api_key='your_api_key_here')\n"
                        "To get an API key, install Qodo CLI via: https://docs.qodo.ai/qodo-documentation/qodo-gen-cli/getting-started/setup-and-quickstart"
                    )
                elif response.status_code != 200:
                    raise exceptions.FailedToGenerateResponseError(f"HTTP {response.status_code}: {response.text}")

                streaming_text = ""
                processed_stream = sanitize_stream(
                    data=response.iter_content(chunk_size=None),
                    intro_value="",
                    to_json=True,
                    skip_markers=["[DONE]"],
                    content_extractor=self._qodo_extractor,
                    yield_raw_on_error=True,
                    raw=raw
                )
                for content_chunk in processed_stream:
                    if content_chunk:
                        yield content_chunk if raw else {"text": content_chunk}
                        if not raw:
                            streaming_text += content_chunk

                self.last_response = {"text": streaming_text}
                self.conversation.update_chat_history(prompt, streaming_text)
            except CurlError as e:
                raise exceptions.FailedToGenerateResponseError(f"Request failed (CurlError): {e}")
            except Exception as e:
                raise exceptions.FailedToGenerateResponseError(f"An unexpected error occurred ({type(e).__name__}): {e}")

        def for_non_stream():
            try:
                payload["stream"] = False
                response = self.session.post(
                    self.url,
                    json=payload,
                    timeout=self.timeout,
                    impersonate=self.fingerprint.get("browser_type", "chrome110")
                )
                # Check for internal server error with session ID in the response
                if response.status_code == 500 and response.text and "Internal server error, session ID:" in response.text:
                    self.url = "https://api.cli.qodo.ai/v2/agentic/continue-task"
                    response = self.session.post(
                        self.url,
                        json=payload,
                        timeout=self.timeout,
                        impersonate=self.fingerprint.get("browser_type", "chrome110")
                    )
                if response.status_code == 401:
                    raise exceptions.FailedToGenerateResponseError(
                        "Invalid API key. You need to provide your own API key.\n"
                        "Usage: QodoAI(api_key='your_api_key_here')\n"
                        "To get an API key, install Qodo CLI via: https://docs.qodo.ai/qodo-documentation/qodo-gen-cli/getting-started/setup-and-quickstart"
                    )
                elif response.status_code != 200:
                    raise exceptions.FailedToGenerateResponseError(f"HTTP {response.status_code}: {response.text}")

                response_text = response.text
                processed_stream = sanitize_stream(
                    data=response_text.splitlines(),
                    intro_value=None,
                    to_json=True,
                    content_extractor=self._qodo_extractor,
                    yield_raw_on_error=True,
                    raw=raw
                )
                full_response = ""
                for content in processed_stream:
                    if content:
                        full_response += content

                self.last_response = {"text": full_response}
                self.conversation.update_chat_history(prompt, full_response)
                return {"text": full_response} if not raw else full_response
            except CurlError as e:
                raise exceptions.FailedToGenerateResponseError(f"Request failed (CurlError): {e}")
            except Exception as e:
                raise exceptions.FailedToGenerateResponseError(f"Request failed ({type(e).__name__}): {e}")

        return for_stream() if stream else for_non_stream()

    def chat(
        self,
        prompt: str,
        stream: bool = False,
        optimizer: Optional[str] = None,
        conversationally: bool = False,
        raw: bool = False,
        **kwargs: Any,
    ) -> Union[str, Generator[str, None, None]]:
        def for_stream():
            for response in self.ask(
                prompt, True, raw=raw, optimizer=optimizer, conversationally=conversationally
            ):
                if raw:
                    yield response
                else:
                    yield self.get_message(response)

        def for_non_stream():
            result = self.ask(
                prompt, False, raw=raw, optimizer=optimizer, conversationally=conversationally
            )
            if raw:
                return result
            else:
                return self.get_message(result)

        return for_stream() if stream else for_non_stream()

    def get_message(self, response: Response) -> str:
        if isinstance(response, str):
            return response
        if isinstance(response, dict):
            text = dict(response).get("text", "")
            return text.replace('\\n', '\n').replace('\\n\\n', '\n\n')
        return str(response)

    def _get_session_id(self) -> str:
        """Get session ID from Qodo API."""
        try:
            # Create temporary session for the info request
            temp_session = Session()
            temp_headers = {
                "Accept": "text/plain",
                "Accept-Encoding": "gzip, deflate, br",
                "Authorization": f"Bearer {self.api_key}",
                "Connection": "close",
                "Content-Type": "application/json",
                "host": "api.cli.qodo.ai",
                "Request-id": str(uuid.uuid4()),
                "User-Agent": self.fingerprint["user_agent"] if hasattr(self, 'fingerprint') else "axios/1.10.0",
            }
            temp_session.headers.update(temp_headers)

            response = temp_session.get(
                self.info_url,
                timeout=self.timeout if hasattr(self, 'timeout') else 30,
                impersonate="chrome110"
            )

            if response.status_code == 200:
                data = response.json()
                session_id = data.get("session-id")
                if session_id:
                    return session_id
            elif response.status_code == 401:
                # API key is invalid
                raise exceptions.FailedToGenerateResponseError(
                    "Invalid API key. You need to provide your own API key.\n"
                    "Usage: QodoAI(api_key='your_api_key_here')\n"
                    "To get an API key, install Qodo CLI via: https://docs.qodo.ai/qodo-documentation/qodo-gen-cli/getting-started/setup-and-quickstart"
                )
            else:
                # Other HTTP errors
                raise exceptions.FailedToGenerateResponseError(
                    f"Failed to authenticate with Qodo API (HTTP {response.status_code}). "
                    "You may need to provide your own API key.\n"
                    "Usage: QodoAI(api_key='your_api_key_here')\n"
                    "To get an API key, install Qodo CLI via: https://docs.qodo.ai/qodo-documentation/qodo-gen-cli/getting-started/setup-and-quickstart"
                )

            # Fallback to generated session ID if API call fails
            from datetime import datetime
            today = datetime.now().strftime("%Y%m%d")
            return f"{today}-{str(uuid.uuid4())}"

        except exceptions.FailedToGenerateResponseError:
            # Re-raise our custom exceptions
            raise
        except Exception as e:
            # For other errors, show the API key message
            raise exceptions.FailedToGenerateResponseError(
                f"Failed to connect to Qodo API: {e}\n"
                "You may need to provide your own API key.\n"
                "Usage: QodoAI(api_key='your_api_key_here')\n"
                "To get an API key, install Qodo CLI via: https://docs.qodo.ai/qodo-documentation/qodo-gen-cli/getting-started/setup-and-quickstart"
            )

    def refresh_session(self):
        """
        Refreshes the session ID by calling the Qodo API.

        Returns:
            str: The new session ID
        """
        self.session_id = self._get_session_id()

        # Update headers with new session ID
        self.headers["Session-id"] = self.session_id
        self.session.headers["Session-id"] = self.session_id

        return self.session_id

    def get_available_models(self) -> Dict[str, Any]:
        """
        Get available models and info from Qodo API.

        Returns:
            Dict containing models, default_model, version, and session info
        """
        try:
            response = self.session.get(
                self.info_url,
                timeout=self.timeout,
                impersonate=self.fingerprint.get("browser_type", "chrome110")
            )

            if response.status_code == 200:
                return response.json()
            elif response.status_code == 401:
                raise exceptions.FailedToGenerateResponseError(
                    "Invalid API key. You need to provide your own API key.\n"
                    "Usage: QodoAI(api_key='your_api_key_here')\n"
                    "To get an API key, install Qodo CLI via: https://docs.qodo.ai/qodo-documentation/qodo-gen-cli/getting-started/setup-and-quickstart"
                )
            else:
                raise exceptions.FailedToGenerateResponseError(f"Failed to get models: HTTP {response.status_code}")

        except CurlError as e:
            raise exceptions.FailedToGenerateResponseError(f"Request failed (CurlError): {e}")
        except Exception as e:
            raise exceptions.FailedToGenerateResponseError(f"Failed to get models ({type(e).__name__}): {e}")


if __name__ == "__main__":
    ai = QodoAI() # u will need to give your API key here to get api install qodo cli via https://docs.qodo.ai/qodo-documentation/qodo-gen-cli/getting-started/setup-and-quickstart
    response = ai.chat("write a poem about india", raw=False, stream=True)
    for chunk in response:
        print(chunk, end='', flush=True)



import re
from typing import Any, Dict, Generator, Optional, Union, cast

from curl_cffi import CurlError
from curl_cffi.requests import Session

from webscout import exceptions
from webscout.AIbase import Provider, Response
from webscout.AIutel import (  # Import sanitize_stream
    AwesomePrompts,
    Conversation,
    Optimizers,
    sanitize_stream,
)
from webscout.litagent import LitAgent


class TurboSeek(Provider):
    """
    This class provides methods for interacting with the TurboSeek API.
    """

    required_auth = False
    AVAILABLE_MODELS = ["Llama 3.1 70B"]

    def __init__(
        self,
        is_conversation: bool = True,
        max_tokens: int = 600,
        timeout: int = 30,
        intro: Optional[str] = None,
        filepath: Optional[str] = None,
        update_file: bool = True,
        proxies: dict = {},
        history_offset: int = 10250,
        act: Optional[str] = None,
        model: str = "Llama 3.1 70B",  # Note: model parameter is not used by the API endpoint
    ):
        """Instantiates TurboSeek

        Args:
            is_conversation (bool, optional): Flag for chatting conversationally. Defaults to True.
            max_tokens (int, optional): Maximum number of tokens to be generated upon completion. Defaults to 600.
            timeout (int, optional): Http request timeout. Defaults to 30.
            intro (str, optional): Conversation introductory prompt. Defaults to None.
            filepath (str, optional): Path to file containing conversation history. Defaults to None.
            update_file (bool, optional): Add new prompts and responses to the file. Defaults to True.
            proxies (dict, optional): Http request proxies. Defaults to {}.
            history_offset (int, optional): Limit conversation history to this number of last texts. Defaults to 10250.
            act (str|int, optional): Awesome prompt key or index. (Used as intro). Defaults to None.
        """
        # Initialize curl_cffi Session
        self.session = Session()
        self.is_conversation = is_conversation
        self.max_tokens_to_sample = max_tokens
        self.chat_endpoint = "https://www.turboseek.io/api/getAnswer"
        self.stream_chunk_size = 64
        self.timeout = timeout
        self.last_response = {}
        self.headers = {
            "accept": "*/*",
            "accept-encoding": "gzip, deflate, br, zstd",
            "accept-language": "en-US,en;q=0.9,en-IN;q=0.8",
            "content-type": "application/json",
            "dnt": "1",
            "origin": "https://www.turboseek.io",
            "priority": "u=1, i",
            "referer": "https://www.turboseek.io/",
            "sec-ch-ua": '"Chromium";v="140", "Not=A?Brand";v="24", "Microsoft Edge";v="140"',
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": '"Windows"',
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin",
            "sec-gpc": "1",
            "user-agent": LitAgent().random(),
        }

        self.__available_optimizers = (
            method
            for method in dir(Optimizers)
            if callable(getattr(Optimizers, method)) and not method.startswith("__")
        )
        # Update curl_cffi session headers and proxies
        self.session.headers.update(self.headers)
        if proxies:
            self.session.proxies.update(proxies)  # Assign proxies directly
        self.conversation = Conversation(
            is_conversation, self.max_tokens_to_sample, filepath, update_file
        )
        self.conversation.history_offset = history_offset

        if act:
            self.conversation.intro = (
                AwesomePrompts().get_act(
                    cast(Union[str, int], act),
                    default=self.conversation.intro,
                    case_insensitive=True,
                )
                or self.conversation.intro
            )
        elif intro:
            self.conversation.intro = intro

    @staticmethod
    def _html_to_markdown(text: str) -> str:
        """Convert basic HTML tags to Markdown."""
        if not text:
            return ""

        # Unescape HTML entities first
        import html

        text = html.unescape(text)

        # Headers
        text = re.sub(r"<h[1-6][^>]*>(.*?)</h[1-6]>", r"\n# \1\n", text)

        # Lists
        text = re.sub(r"<li[^>]*>(.*?)</li>", r"\n* \1", text)
        text = re.sub(r"<(ul|ol)[^>]*>", r"\n", text)
        text = re.sub(r"</(ul|ol)>", r"\n", text)

        # Paragraphs and Breaks
        text = re.sub(r"</p>", r"\n\n", text)
        text = re.sub(r"<p[^>]*>", r"\n", text)
        text = re.sub(r"<br\s*/?>", r"\n", text)

        # Bold and Italic
        text = re.sub(r"<(strong|b)[^>]*>(.*?)</\1>", r"**\2**", text)
        text = re.sub(r"<(em|i)[^>]*>(.*?)</\1>", r"*\2*", text)

        # Remove structural tags
        text = re.sub(
            r"</?(section|div|span|article|header|footer)[^>]*>", "", text, flags=re.IGNORECASE
        )

        # Final cleanup of remaining tags
        text = re.sub(r"<[^>]*>", "", text)

        return text

    @staticmethod
    def _turboseek_extractor(chunk: Any) -> Optional[str]:
        """Extracts content from TurboSeek stream."""
        if isinstance(chunk, str):
            # The API now returns raw HTML chunks
            return chunk
        return None

    def ask(
        self,
        prompt: str,
        stream: bool = False,
        raw: bool = False,
        optimizer: Optional[str] = None,
        conversationally: bool = False,
        **kwargs: Any,
    ) -> Response:
        """Chat with AI"""
        conversation_prompt = self.conversation.gen_complete_prompt(prompt)
        if optimizer:
            if optimizer in self.__available_optimizers:
                conversation_prompt = getattr(Optimizers, optimizer)(
                    conversation_prompt if conversationally else prompt
                )
            else:
                raise Exception(f"Optimizer is not one of {self.__available_optimizers}")

        payload = {"question": conversation_prompt, "sources": []}

        def for_stream():
            try:
                response = self.session.post(
                    self.chat_endpoint,
                    json=payload,
                    stream=True,
                    timeout=self.timeout,
                    impersonate="chrome120",
                )
                if not response.ok:
                    raise exceptions.FailedToGenerateResponseError(
                        f"Failed to generate response - ({response.status_code}, {response.reason}) - {response.text}"
                    )

                streaming_text = ""
                # The API returns raw HTML chunks now, no "data:" prefix
                processed_stream = sanitize_stream(
                    data=response.iter_content(chunk_size=None),
                    intro_value=None,
                    to_json=False,
                    strip_chars="",  # Disable default lstrip to preserve spacing
                    content_extractor=self._turboseek_extractor,
                    yield_raw_on_error=True,
                    raw=raw,
                )

                for content_chunk in processed_stream:
                    if content_chunk is None:
                        continue

                    if raw:
                        yield content_chunk
                    else:
                        if isinstance(content_chunk, str):
                            # In streaming mode, stripping HTML incrementally is hard.
                            # We'll just yield the chunk but clean it slightly.
                            # For full Markdown conversion, use non-streaming or aggregate it.
                            clean_chunk = re.sub(r"<[^>]*>", "", content_chunk)
                            if clean_chunk:
                                streaming_text += clean_chunk
                                self.last_response.update(dict(text=streaming_text))
                                yield dict(text=clean_chunk)

                if not raw and streaming_text:
                    self.conversation.update_chat_history(prompt, streaming_text)
            except CurlError as e:
                raise exceptions.FailedToGenerateResponseError(f"Request failed (CurlError): {e}")
            except Exception as e:
                raise exceptions.FailedToGenerateResponseError(
                    f"An unexpected error occurred ({type(e).__name__}): {e}"
                )

        def for_non_stream():
            full_html = ""
            try:
                # Iterate over the stream in raw mode to get full HTML
                # We use ask(..., raw=True) internally or just the local for_stream
                # Actually, let's just make a sub-call
                response = self.session.post(
                    self.chat_endpoint, json=payload, timeout=self.timeout, impersonate="chrome120"
                )
                full_html = response.text
            except Exception as e:
                raise exceptions.FailedToGenerateResponseError(
                    f"Failed to get non-stream response: {e}"
                ) from e

            # Convert full HTML to Markdown
            final_text = self._html_to_markdown(full_html).strip()
            self.last_response = {"text": final_text}
            return self.last_response

        return for_stream() if stream else for_non_stream()

    def chat(
        self,
        prompt: str,
        stream: bool = False,
        optimizer: Optional[str] = None,
        conversationally: bool = False,
        **kwargs: Any,
    ) -> Union[str, Generator[str, None, None]]:
        """Generate response `str`
        Args:
            prompt (str): Prompt to be send.
            stream (bool, optional): Flag for streaming response. Defaults to False.
            optimizer (str, optional): Prompt optimizer name - `[code, shell_command]`. Defaults to None.
            conversationally (bool, optional): Chat conversationally when using optimizer. Defaults to False.
            **kwargs: Additional parameters including raw.
        Returns:
            str: Response generated
        """
        raw = kwargs.get("raw", False)

        def for_stream():
            for response in self.ask(
                prompt, True, raw=raw, optimizer=optimizer, conversationally=conversationally
            ):
                if raw:
                    yield cast(str, response)
                else:
                    yield self.get_message(cast(Response, response))

        def for_non_stream():
            result = self.ask(
                prompt,
                False,
                raw=raw,
                optimizer=optimizer,
                conversationally=conversationally,
            )
            if raw:
                return cast(str, result)
            else:
                return self.get_message(result)

        return for_stream() if stream else for_non_stream()

    def get_message(self, response: Response) -> str:
        """Retrieves message only from response

        Args:
            response (Response): Response generated by `self.ask`

        Returns:
            str: Message extracted
        """
        if not isinstance(response, dict):
            return str(response)
        response_dict = cast(Dict[str, Any], response)
        return response_dict.get("text", "")


if __name__ == "__main__":
    import sys

    ai = TurboSeek(timeout=60)

    # helper for safe printing on windows
    def safe_print(text, end="\n"):
        try:
            sys.stdout.write(text + end)
        except UnicodeEncodeError:
            sys.stdout.write(text.encode("ascii", "ignore").decode("ascii") + end)
        sys.stdout.flush()

    safe_print("\n=== Testing Non-Streaming ===")
    response = ai.chat("How can I get a 6 pack in 3 months?", stream=False)
    if isinstance(response, str):
        safe_print(response)
    else:
        safe_print(str(response))

    safe_print("\n=== Testing Streaming ===")
    stream_resp = ai.chat("How can I get a 6 pack in 3 months?", stream=True)
    if hasattr(stream_resp, "__iter__") and not isinstance(stream_resp, (str, bytes)):
        for chunk in stream_resp:
            safe_print(chunk, end="")
    else:
        safe_print(str(stream_resp))
    safe_print("")



import json
import time
import uuid
from typing import Any, Dict, Generator, List, Optional, Union, cast

import requests

from webscout.Provider.Openai_comp.base import (
    BaseChat,
    BaseCompletions,
    OpenAICompatibleProvider,
    SimpleModelList,
)
from webscout.Provider.Openai_comp.utils import (
    ChatCompletion,
    ChatCompletionChunk,
    ChatCompletionMessage,
    Choice,
    ChoiceDelta,
    CompletionUsage,
)

from ...litagent import LitAgent


class Completions(BaseCompletions):
    def __init__(self, client: "Zenmux"):
        self._client = client

    def create(
        self,
        *,
        model: str,
        messages: List[Dict[str, Any]],
        max_tokens: Optional[int] = 1024,
        stream: bool = False,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        timeout: Optional[int] = None,
        proxies: Optional[Dict[str, str]] = None,
        tools: Optional[List[Union[Dict[str, Any], Any]]] = None,
        tool_choice: Optional[Union[str, Dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> Union[ChatCompletion, Generator[ChatCompletionChunk, None, None]]:
        payload: Dict[str, Any] = {
            "messages": messages,
            "model": model,
            "stream": stream,
        }
        if stream:
            payload["stream_options"] = {"include_usage": True}
        if max_tokens is not None:
            payload["max_tokens"] = max_tokens
        if temperature is not None:
            payload["temperature"] = temperature
        if top_p is not None:
            payload["top_p"] = top_p
        if tools:
            payload["tools"] = self.format_tool_calls(tools)
        payload.update(kwargs)
        request_id = f"chatcmpl-{uuid.uuid4()}"
        created_time = int(time.time())
        if stream:
            return self._create_stream(request_id, created_time, model, payload, timeout, proxies)
        else:
            return self._create_non_stream(
                request_id, created_time, model, payload, timeout, proxies
            )

    def _create_stream(
        self,
        request_id: str,
        created_time: int,
        model: str,
        payload: Dict[str, Any],
        timeout: Optional[int] = None,
        proxies: Optional[Dict[str, str]] = None,
    ) -> Generator[ChatCompletionChunk, None, None]:
        try:
            response = self._client.session.post(
                self._client.base_url,
                headers=self._client.headers,
                json=payload,
                stream=True,
                timeout=timeout or self._client.timeout,
                proxies=proxies,
            )
            response.raise_for_status()
            prompt_tokens = 0
            completion_tokens = 0
            total_tokens = 0
            for line in response.iter_lines(decode_unicode=True):
                if not line:
                    continue
                if isinstance(line, bytes):
                    try:
                        line = line.decode("utf-8")
                    except Exception:
                        continue
                if line.startswith("data: "):
                    json_str = line[6:]
                else:
                    json_str = line
                if json_str == "[DONE]":
                    break
                try:
                    data = json.loads(json_str)
                    choice_data = data.get("choices", [{}])[0]
                    delta_data = choice_data.get("delta", {})
                    finish_reason = choice_data.get("finish_reason")
                    usage_data = data.get("usage", {})
                    if usage_data:
                        prompt_tokens = usage_data.get("prompt_tokens", prompt_tokens)
                        completion_tokens = usage_data.get("completion_tokens", completion_tokens)
                        total_tokens = usage_data.get("total_tokens", total_tokens)
                    content_piece = None
                    role = None
                    tool_calls = None
                    if delta_data:
                        content_piece = delta_data.get("content") or delta_data.get("text")
                        role = delta_data.get("role")
                        tool_calls = delta_data.get("tool_calls")
                    else:
                        message_d = choice_data.get("message", {})
                        role = message_d.get("role")
                        content_piece = message_d.get("content") or message_d.get("text")
                        tool_calls = message_d.get("tool_calls")
                    if content_piece:
                        completion_tokens += 1
                        total_tokens = prompt_tokens + completion_tokens
                    delta = ChoiceDelta(content=content_piece, role=role, tool_calls=tool_calls)
                    choice = Choice(
                        index=choice_data.get("index", 0),
                        delta=delta,
                        finish_reason=finish_reason,
                        logprobs=choice_data.get("logprobs"),
                    )
                    chunk = ChatCompletionChunk(
                        id=request_id,
                        choices=[choice],
                        created=created_time,
                        model=model,
                        system_fingerprint=data.get("system_fingerprint"),
                    )
                    chunk.usage = {
                        "prompt_tokens": prompt_tokens,
                        "completion_tokens": completion_tokens,
                        "total_tokens": total_tokens,
                        "estimated_cost": None,
                    }
                    yield chunk
                except json.JSONDecodeError:
                    continue
            delta = ChoiceDelta(content=None, role=None, tool_calls=None)
            choice = Choice(index=0, delta=delta, finish_reason="stop", logprobs=None)
            chunk = ChatCompletionChunk(
                id=request_id,
                choices=[choice],
                created=created_time,
                model=model,
                system_fingerprint=None,
            )
            chunk.usage = {
                "prompt_tokens": prompt_tokens,
                "completion_tokens": completion_tokens,
                "total_tokens": total_tokens,
                "estimated_cost": None,
            }
            yield chunk
        except Exception as e:
            print(f"Error during Zenmux stream request: {e}")
            raise IOError(f"Zenmux request failed: {e}") from e

    def _create_non_stream(
        self,
        request_id: str,
        created_time: int,
        model: str,
        payload: Dict[str, Any],
        timeout: Optional[int] = None,
        proxies: Optional[Dict[str, str]] = None,
    ) -> ChatCompletion:
        try:
            response = self._client.session.post(
                self._client.base_url,
                headers=self._client.headers,
                json=payload,
                timeout=timeout or self._client.timeout,
                proxies=proxies,
            )
            response.raise_for_status()
            data = response.json()
            choices_data = data.get("choices", [])
            usage_data = data.get("usage", {})
            choices = []
            for choice_d in choices_data:
                message_d = choice_d.get("message")
                if not message_d and "delta" in choice_d:
                    delta = choice_d["delta"]
                    message_d = {
                        "role": delta.get("role", "assistant"),
                        "content": delta.get("content", ""),
                    }
                if not message_d:
                    message_d = {"role": "assistant", "content": ""}
                message = ChatCompletionMessage(
                    role=message_d.get("role", "assistant"), content=message_d.get("content", "")
                )
                choice = Choice(
                    index=choice_d.get("index", 0),
                    message=message,
                    finish_reason=choice_d.get("finish_reason", "stop"),
                )
                choices.append(choice)
            usage = CompletionUsage(
                prompt_tokens=usage_data.get("prompt_tokens", 0),
                completion_tokens=usage_data.get("completion_tokens", 0),
                total_tokens=usage_data.get("total_tokens", 0),
            )
            completion = ChatCompletion(
                id=request_id,
                choices=choices,
                created=created_time,
                model=data.get("model", model),
                usage=usage,
            )
            return completion
        except Exception as e:
            print(f"Error during Zenmux non-stream request: {e}")
            raise IOError(f"Zenmux request failed: {e}") from e


class Chat(BaseChat):
    def __init__(self, client: "Zenmux"):
        self.completions = Completions(client)


class Zenmux(OpenAICompatibleProvider):
    required_auth = True
    AVAILABLE_MODELS = [
        "z-ai/glm-4.6v-flash",
    ]

    def __init__(self, browser: str = "chrome", api_key: Optional[str] = None):
        self.timeout = None
        self.base_url = "https://zenmux.ai/api/v1/chat/completions"
        self.session = requests.Session()
        agent = LitAgent()
        fingerprint = agent.generate_fingerprint(browser)
        self.headers = {
            "Accept": fingerprint["accept"],
            "Accept-Language": fingerprint["accept_language"],
            "Content-Type": "application/json",
            "DNT": "1",
            "Origin": "https://zenmux.ai",
            "Referer": "https://zenmux.ai/",
            "Priority": "u=1, i",
            "Sec-CH-UA": fingerprint.get("sec_ch_ua")
            or '"Not)A;Brand";v="99", "Microsoft Edge";v="127", "Chromium";v="127"',
            "Sec-CH-UA-Mobile": "?0",
            "Sec-CH-UA-Platform": f'"{fingerprint["platform"]}"',
            "Sec-Fetch-Dest": "empty",
            "Sec-Fetch-Mode": "cors",
            "Sec-Fetch-Site": "same-origin",
            "Sec-GPC": "1",
            "User-Agent": fingerprint.get("user_agent"),
        }
        if api_key is not None:
            self.headers["Authorization"] = f"Bearer {api_key}"
        self.session.headers.update(self.headers)
        self.chat = Chat(self)
        try:
            self.update_available_models(api_key)
        except Exception:
            pass

    @classmethod
    def get_models(cls, api_key: Optional[str] = None):
        try:
            headers = {
                "Accept": "application/json",
                "Content-Type": "application/json",
                "User-Agent": "Mozilla/5.0",
            }
            try:
                from curl_cffi.requests import Session as CurlSession

                curl_available = True
            except Exception:
                CurlSession = None  # type: ignore
                curl_available = False
            try:
                from ...litagent import LitAgent

                agent = LitAgent()
                fingerprint = agent.generate_fingerprint("chrome")
                headers.update(
                    {
                        "Accept": fingerprint.get("accept", headers["Accept"]),
                        "Accept-Language": fingerprint.get("accept_language", "en-US,en;q=0.9"),
                        "User-Agent": fingerprint.get("user_agent", headers["User-Agent"]),
                        "Sec-CH-UA": fingerprint.get("sec_ch_ua")
                        or '"Not)A;Brand";v="99", "Microsoft Edge";v="127", "Chromium";v="127"',
                        "Sec-CH-UA-Mobile": "?0",
                        "Sec-CH-UA-Platform": f'"{fingerprint.get("platform", "Windows")}"',
                    }
                )
            except Exception:
                pass
            if api_key:
                headers["Authorization"] = f"Bearer {api_key}"
            if curl_available and CurlSession is not None:
                session = CurlSession()
                response = session.get(
                    "https://zenmux.ai/api/v1/models",
                    headers=headers,
                    impersonate="chrome110",
                    timeout=10,
                )
            else:
                response = requests.get(
                    "https://zenmux.ai/api/v1/models", headers=headers, timeout=10
                )
            if getattr(response, "status_code", None) and response.status_code != 200:
                return cls.AVAILABLE_MODELS
            data = response.json()

            if isinstance(data, dict) and "data" in data and isinstance(data["data"], list):
                parsed = []
                for entry in data["data"]:
                    if isinstance(entry, str):
                        parsed.append(entry)
                    elif isinstance(entry, dict) and "id" in entry and isinstance(entry["id"], str):
                        parsed.append(entry["id"])
                if parsed:
                    return parsed

            def _parse(data_obj):
                models = []
                if isinstance(data_obj, list):
                    for item in data_obj:
                        if isinstance(item, str):
                            models.append(item)
                        elif isinstance(item, dict):
                            for key in ("id", "model", "name"):
                                if key in item and isinstance(item[key], str):
                                    models.append(item[key])
                                    break
                            else:
                                for v in item.values():
                                    if isinstance(v, str):
                                        models.append(v)
                                        break
                elif isinstance(data_obj, dict):
                    for key in ("data", "models", "result"):
                        if key in data_obj:
                            models.extend(_parse(data_obj[key]))
                    for k, v in data_obj.items():
                        if isinstance(v, dict):
                            models.append(k)
                return models

            models = _parse(data)
            models = [m for m in dict.fromkeys(models) if m]
            return models if models else cls.AVAILABLE_MODELS
        except Exception:
            return cls.AVAILABLE_MODELS

    @classmethod
    def update_available_models(cls, api_key: Optional[str] = None):
        try:
            models = cls.get_models(api_key)
            if models:
                cls.AVAILABLE_MODELS = models
        except Exception:
            pass

    @property
    def models(self) -> SimpleModelList:
        return SimpleModelList(self.AVAILABLE_MODELS)



from typing import Any, Dict, Generator, Optional, Union, cast
from uuid import uuid4

from curl_cffi import CurlError

# Import HTTPVersion enum
from curl_cffi.const import CurlHttpVersion
from curl_cffi.requests import Session

from webscout import exceptions
from webscout.AIbase import Provider, Response
from webscout.AIutel import (  # Import sanitize_stream
    AwesomePrompts,
    Conversation,
    Optimizers,
    sanitize_stream,
)
from webscout.litagent import LitAgent


class X0GPT(Provider):
    """
    A class to interact with the x0-gpt.devwtf.in API.

    Attributes:
        system_prompt (str): The system prompt to define the assistant's role.

    Examples:
        >>> from webscout.Provider.x0gpt import X0GPT
        >>> ai = X0GPT()
        >>> response = ai.chat("What's the weather today?")
        >>> print(response)
        'The weather today is sunny with a high of 75°F.'
    """

    required_auth = False
    AVAILABLE_MODELS = ["UNKNOWN"]

    def __init__(
        self,
        is_conversation: bool = True,
        max_tokens: int = 600,
        timeout: int = 30,
        intro: Optional[str] = None,
        filepath: Optional[str] = None,
        update_file: bool = True,
        proxies: dict = {},
        history_offset: int = 10250,
        act: Optional[str] = None,
        system_prompt: str = "You are a helpful assistant.",
        model: str = "UNKNOWN",
    ):
        """
        Initializes the X0GPT API with given parameters.

        Args:
            is_conversation (bool): Whether the provider is in conversation mode.
            max_tokens (int): Maximum number of tokens to sample.
            timeout (int): Timeout for API requests.
            intro (str): Introduction message for the conversation.
            filepath (str): Filepath for storing conversation history.
            update_file (bool): Whether to update the conversation history file.
            proxies (dict): Proxies for the API requests.
            history_offset (int): Offset for conversation history.
            act (str): Act for the conversation.
            system_prompt (str): The system prompt to define the assistant's role.

        Examples:
            >>> ai = X0GPT(system_prompt="You are a friendly assistant.")
            >>> print(ai.system_prompt)
            'You are a friendly assistant.'
        """
        # Initialize curl_cffi Session instead of requests.Session
        self.session = Session()
        self.is_conversation = is_conversation
        self.max_tokens_to_sample = max_tokens
        self.api_endpoint = "https://x0-gpt.devwtf.in/api/stream/reply"
        self.timeout = timeout
        self.last_response = {}
        self.system_prompt = system_prompt

        # Initialize LitAgent for user agent generation
        self.agent = LitAgent()

        self.headers = {
            "authority": "x0-gpt.devwtf.in",
            "method": "POST",
            "path": "/api/stream/reply",
            "scheme": "https",
            "accept": "*/*",
            "accept-encoding": "gzip, deflate, br, zstd",  # Keep zstd for now
            "accept-language": "en-US,en;q=0.9,en-IN;q=0.8",
            # "content-length": "114", # Let curl_cffi handle content-length
            "content-type": "application/json",
            "dnt": "1",
            "origin": "https://x0-gpt.devwtf.in",
            # "priority": "u=1, i", # Remove priority header
            "referer": "https://x0-gpt.devwtf.in/chat",
            "sec-ch-ua": '"Not)A;Brand";v="99", "Microsoft Edge";v="127", "Chromium";v="127"',
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": '"Windows"',
            "user-agent": self.agent.random(),
        }

        self.__available_optimizers = (
            method
            for method in dir(Optimizers)
            if callable(getattr(Optimizers, method)) and not method.startswith("__")
        )
        # Update curl_cffi session headers and proxies
        self.session = Session()
        self.session.headers.update(self.headers)
        if proxies:
            self.session.proxies.update(proxies)

        self.conversation = Conversation(
            is_conversation, self.max_tokens_to_sample, filepath, update_file
        )
        self.conversation.history_offset = history_offset

        if act:
            self.conversation.intro = (
                AwesomePrompts().get_act(
                    cast(Union[str, int], act),
                    default=self.conversation.intro,
                    case_insensitive=True,
                )
                or self.conversation.intro
            )
        elif intro:
            self.conversation.intro = intro

    def ask(
        self,
        prompt: str,
        stream: bool = False,
        raw: bool = False,
        optimizer: Optional[str] = None,
        conversationally: bool = False,
        **kwargs: Any,
    ) -> Response:
        """
        Sends a prompt to the x0-gpt.devwtf.in API and returns the response.

        Args:
            prompt (str): The prompt to send to the API.
            stream (bool): Whether to stream the response.
            raw (bool): Whether to return the raw response.
            optimizer (str): Optimizer to use for the prompt.
            conversationally (bool): Whether to generate the prompt conversationally.

        Returns:
            Dict[str, Any]: The API response.

        Examples:
            >>> ai = X0GPT()
            >>> response = ai.ask("Tell me a joke!")
            >>> print(response)
            {'text': 'Why did the scarecrow win an award? Because he was outstanding in his field!'}
        """
        conversation_prompt = self.conversation.gen_complete_prompt(prompt)
        if optimizer:
            if optimizer in self.__available_optimizers:
                conversation_prompt = getattr(Optimizers, optimizer)(
                    conversation_prompt if conversationally else prompt
                )
            else:
                raise Exception(f"Optimizer is not one of {self.__available_optimizers}")

        payload = {
            "messages": [
                {"role": "system", "content": self.system_prompt},
                {"role": "user", "content": conversation_prompt},
            ],
            "chatId": uuid4().hex,
            "namespace": None,
        }

        def for_stream():
            try:
                # Use curl_cffi session post with updated impersonate and http_version
                response = self.session.post(
                    self.api_endpoint,
                    headers=self.headers,
                    json=payload,
                    stream=True,
                    timeout=self.timeout,
                    impersonate="chrome120",  # Try a different impersonation profile
                    http_version=CurlHttpVersion.V1_1,  # Force HTTP/1.1
                )
                if not response.ok:
                    raise exceptions.FailedToGenerateResponseError(
                        f"Failed to generate response - ({response.status_code}, {response.reason}) - {response.text}"
                    )

                streaming_response = ""
                # Use sanitize_stream with regex-based extraction and filtering
                processed_stream = sanitize_stream(
                    data=response.iter_content(chunk_size=None),  # Pass byte iterator
                    intro_value=None,  # No simple prefix to remove here
                    to_json=False,  # Content is not JSON
                    # Use regex to extract content from x0gpt format '0:"..."'
                    extract_regexes=[r'0:"(.*?)"'],
                    # Skip empty chunks, connection status messages, and control characters
                    skip_regexes=[
                        r"^\s*$",  # Empty lines
                        r"data:\s*\[DONE\]",  # Stream end markers
                        r"event:\s*",  # SSE event headers
                        r"^\d+:\s*$",  # Standalone numbers
                        r"^:\s*$",  # Colon-only lines
                        r"^\s*[\x00-\x1f]+\s*$",  # Control characters
                    ],
                    raw=raw,
                )

                for content_chunk in processed_stream:
                    # Always yield as string, even in raw mode
                    if isinstance(content_chunk, bytes):
                        content_chunk = content_chunk.decode("utf-8", errors="ignore")

                    if raw:
                        yield content_chunk
                    else:
                        if content_chunk and isinstance(content_chunk, str):
                            # Handle unicode escapes and clean up the content
                            try:
                                # Decode unicode escapes like \u00e9
                                clean_content = content_chunk.encode().decode("unicode_escape")
                                # Handle escaped backslashes and quotes
                                clean_content = clean_content.replace("\\\\", "\\").replace(
                                    '\\"', '"'
                                )
                                streaming_response += clean_content
                                yield dict(text=clean_content)
                            except (UnicodeDecodeError, UnicodeEncodeError):
                                # Fallback to original content if unicode processing fails
                                streaming_response += content_chunk
                                yield dict(text=content_chunk)

                self.last_response.update(dict(text=streaming_response))
                self.conversation.update_chat_history(prompt, self.get_message(self.last_response))
            except CurlError as e:  # Catch CurlError
                raise exceptions.FailedToGenerateResponseError(f"Request failed (CurlError): {e}")
            except Exception as e:  # Catch other potential exceptions
                # Include the original exception type in the message for clarity
                raise exceptions.FailedToGenerateResponseError(
                    f"An unexpected error occurred ({type(e).__name__}): {e}"
                )

        def for_non_stream():
            # This function implicitly uses the updated for_stream
            if stream:
                return for_stream()
            for _ in for_stream():
                pass
            return self.last_response

        return for_stream() if stream else for_non_stream()

    def chat(
        self,
        prompt: str,
        stream: bool = False,
        optimizer: Optional[str] = None,
        conversationally: bool = False,
        **kwargs: Any,
    ) -> Union[str, Generator[str, None, None]]:
        """
        Generates a response from the X0GPT API.

        Args:
            prompt (str): The prompt to send to the API.
            stream (bool): Whether to stream the response.
            optimizer (str): Optimizer to use for the prompt.
            conversationally (bool): Whether to generate the prompt conversationally.
            **kwargs: Additional parameters including raw.

        Returns:
            str: The API response.

        Examples:
            >>> ai = X0GPT()
            >>> response = ai.chat("What's the weather today?")
            >>> print(response)
            'The weather today is sunny with a high of 75°F.'
        """
        raw = kwargs.get("raw", False)

        def for_stream():
            for response in self.ask(
                prompt, True, raw=raw, optimizer=optimizer, conversationally=conversationally
            ):
                if raw:
                    yield cast(str, response)
                else:
                    yield self.get_message(cast(Dict[str, Any], response))

        def for_non_stream():
            result = self.ask(
                prompt,
                False,
                raw=raw,
                optimizer=optimizer,
                conversationally=conversationally,
            )
            if raw:
                return cast(str, result)
            else:
                return self.get_message(cast(Dict[str, Any], result))

        return for_stream() if stream else for_non_stream()

    def get_message(self, response: Response) -> str:
        if not isinstance(response, dict):
            return str(response)
        response_dict = cast(Dict[str, Any], response)
        text = response_dict.get("text", "")
        return text


if __name__ == "__main__":
    from rich import print

    ai = X0GPT(timeout=5000)
    # None Stream Test
    response = ai.chat("write a poem about AI", stream=False, raw=False)
    print(response)
    print()
    print("=====================Stream===============================")
    response = ai.chat("write a poem about AI", stream=True, raw=False)
    for chunk in response:
        print(chunk, end="", flush=True)



import base64
import json
from typing import Any, Dict, Generator, Optional, Union, cast

from curl_cffi import CurlError
from curl_cffi.requests import Session

from webscout import exceptions
from webscout.AIbase import Provider, Response
from webscout.AIutel import AwesomePrompts, Conversation, Optimizers, sanitize_stream


class TwoAI(Provider):
    """
    A class to interact with the Two AI API (v2) with LitAgent user-agent.
    SUTRA is a family of large multi-lingual language models (LMLMs) developed by TWO AI.
    SUTRA's dual-transformer extends the power of both MoE and Dense AI language model architectures,
    delivering cost-efficient multilingual capabilities for over 50+ languages.

    API keys must be provided directly by the user.
    """

    required_auth = True
    AVAILABLE_MODELS = [
        "sutra-v2",  # Multilingual AI model for instruction execution and conversational intelligence
        "sutra-r0",  # Advanced reasoning model for complex problem-solving and deep contextual understanding
    ]

    def __init__(
        self,
        api_key: str,
        is_conversation: bool = True,
        max_tokens: int = 1024,
        timeout: int = 30,
        intro: Optional[str] = None,
        filepath: Optional[str] = None,
        update_file: bool = True,
        proxies: dict = {},
        history_offset: int = 10250,
        act: Optional[str] = None,
        model: str = "sutra-v2",  # Default model
        temperature: float = 0.6,
        system_message: str = "You are a helpful assistant."
    ):
        """
        Initializes the TwoAI API client.

        Args:
            api_key: TwoAI API key (required).
            is_conversation: Whether to maintain conversation history.
            max_tokens: Maximum number of tokens to generate.
            timeout: Request timeout in seconds.
            intro: Introduction text for the conversation.
            filepath: Path to save conversation history.
            update_file: Whether to update the conversation history file.
            proxies: Proxy configuration for requests.
            history_offset: Maximum history length in characters.
            act: Persona for the conversation.
            model: Model to use. Must be one of AVAILABLE_MODELS.
            temperature: Temperature for generation (0.0 to 1.0).
            system_message: System message to use for the conversation.
        """
        if model not in self.AVAILABLE_MODELS:
            raise ValueError(f"Invalid model: {model}. Choose from: {self.AVAILABLE_MODELS}")

        if not api_key:
            raise exceptions.AuthenticationError("TwoAI API key is required.")

        self.url = "https://chatsutra-server.account-2b0.workers.dev/v2/chat/completions"  # Correct API endpoint
        self.headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/140.0.0.0 Safari/537.36 Edg/140.0.0.0',
            'Accept': 'application/json',
            'Accept-Encoding': 'gzip, deflate, br, zstd',
            'Accept-Language': 'en-US,en;q=0.9,en-IN;q=0.8',
            'Content-Type': 'application/json',
            'Origin': 'https://chat.two.ai',
            'Referer': 'https://chatsutra-server.account-2b0.workers.dev/',
            'Sec-Ch-Ua': '"Chromium";v="140", "Not=A?Brand";v="24", "Microsoft Edge";v="140"',
            'Sec-Ch-Ua-Mobile': '?0',
            'Sec-Ch-Ua-Platform': '"Windows"',
            'Sec-Fetch-Dest': 'empty',
            'Sec-Fetch-Mode': 'cors',
            'Sec-Fetch-Site': 'cross-site',
            'Sec-Gpc': '1',
            'Dnt': '1',
            'X-Session-Token': api_key  # Using session token instead of Bearer auth
        }

        # Initialize curl_cffi Session
        self.session = Session()
        self.session.headers.update(self.headers)
        if proxies:
            self.session.proxies.update(proxies)

        self.is_conversation = is_conversation
        self.max_tokens_to_sample = max_tokens
        self.timeout = timeout
        self.last_response = {}
        self.model = model
        self.temperature = temperature
        self.system_message = system_message
        self.api_key = api_key

        self.__available_optimizers = (
            method
            for method in dir(Optimizers)
            if callable(getattr(Optimizers, method)) and not method.startswith("__")
        )
        self.conversation = Conversation(
            is_conversation, self.max_tokens_to_sample, filepath, update_file
        )
        self.conversation.history_offset = history_offset

        if act:
            self.conversation.intro = AwesomePrompts().get_act(cast(Union[str, int], act), default=self.conversation.intro, case_insensitive=True
            ) or self.conversation.intro
        elif intro:
            self.conversation.intro = intro

    @staticmethod
    def _twoai_extractor(chunk_json: Union[str, Dict[str, Any]]) -> Optional[str]:
        """Extracts content from TwoAI v2 stream JSON objects."""
        if not isinstance(chunk_json, dict) or "choices" not in chunk_json or not chunk_json["choices"]:
            return None

        delta = chunk_json["choices"][0].get("delta")
        if not isinstance(delta, dict):
            return None

        content = delta.get("content")
        return content if isinstance(content, str) else None

    def encode_image(self, image_path: str) -> str:
        """
        Encode an image file to base64 string.

        Args:
            image_path: Path to the image file

        Returns:
            Base64 encoded string of the image
        """
        with open(image_path, "rb") as image_file:
            return base64.b64encode(image_file.read()).decode('utf-8')

    def ask(
        self,
        prompt: str,
        stream: bool = True,
        raw: bool = False,
        optimizer: Optional[str] = None,
        conversationally: bool = False,
        **kwargs: Any,
    ) -> Union[Dict[str, Any], Generator]:
        online_search = kwargs.get("online_search", True)
        image_path = kwargs.get("image_path")
        conversation_prompt = self.conversation.gen_complete_prompt(prompt)
        if optimizer:
            if optimizer in self.__available_optimizers:
                conversation_prompt = getattr(Optimizers, optimizer)(conversation_prompt if conversationally else prompt)
            else:
                raise Exception(f"Optimizer is not one of {self.__available_optimizers}")

        # Prepare messages with image if provided
        if image_path:
            # Create a message with image content
            image_content = {
                "type": "image_url",
                "image_url": {
                    "url": f"data:image/jpeg;base64,{self.encode_image(image_path)}"
                }
            }
            user_message = {
                "role": "user",
                "content": [
                    {"type": "text", "text": conversation_prompt},
                    image_content
                ]
            }
        else:
            # Text-only message
            user_message = {"role": "user", "content": conversation_prompt}

        # Prepare the payload
        payload = {
            "messages": [
                *([{"role": "system", "content": self.system_message}] if self.system_message else []),
                user_message
            ],
            "model": self.model,
            "temperature": self.temperature,
            "max_tokens": self.max_tokens_to_sample,
            "extra_body": {
                "online_search": online_search,
            }
        }

        def for_stream():
            streaming_text = ""  # Initialize outside try block
            try:
                response = self.session.post(
                    self.url,
                    json=payload,
                    stream=True,
                    timeout=self.timeout
                )

                if response.status_code != 200:
                    error_detail = response.text
                    try:
                        error_json = response.json()
                        error_detail = error_json.get("error", {}).get("message", error_detail)
                    except json.JSONDecodeError:
                        pass
                    raise exceptions.FailedToGenerateResponseError(
                        f"Request failed with status code {response.status_code} - {error_detail}"
                    )

                # Use sanitize_stream to process the SSE stream
                processed_stream = sanitize_stream(
                    data=response.iter_content(chunk_size=None),
                    intro_value="data:",
                    to_json=True,
                    skip_markers=["[DONE]"],
                    content_extractor=self._twoai_extractor,
                    yield_raw_on_error=False
                )

                for content_chunk in processed_stream:
                    if content_chunk and isinstance(content_chunk, str):
                        streaming_text += content_chunk
                        resp = {"text": content_chunk}
                        yield resp if not raw else content_chunk

                # If stream completes successfully, update history
                self.last_response = {"text": streaming_text}
                self.conversation.update_chat_history(prompt, streaming_text)

            except CurlError as e:
                raise exceptions.FailedToGenerateResponseError(f"Request failed (CurlError): {e}") from e
            except exceptions.FailedToGenerateResponseError:
                raise  # Re-raise specific exception
            except Exception as e:
                raise exceptions.FailedToGenerateResponseError(f"An unexpected error occurred during streaming ({type(e).__name__}): {e}") from e
            finally:
                # Ensure history is updated even if stream ends abruptly but text was received
                if streaming_text and not self.last_response:  # Check if last_response wasn't set in the try block
                    self.last_response = {"text": streaming_text}
                    self.conversation.update_chat_history(prompt, streaming_text)


        def for_non_stream():
            # Non-stream still uses the stream internally and aggregates
            streaming_text = ""
            # We need to consume the generator from for_stream()
            gen = for_stream()
            try:
                for chunk_data in gen:
                    if isinstance(chunk_data, dict) and "text" in chunk_data:
                        streaming_text += chunk_data["text"]
                    elif isinstance(chunk_data, str): # Handle raw=True case
                        streaming_text += chunk_data
            except exceptions.FailedToGenerateResponseError:
                 # If the underlying stream fails, re-raise the error
                 raise
            # self.last_response and history are updated within for_stream's try/finally
            return self.last_response # Return the final aggregated dict

        # The API uses SSE streaming for all requests, so we always use streaming
        return for_stream()

    def chat(
        self,
        prompt: str,
        stream: bool = True,
        optimizer: Optional[str] = None,
        conversationally: bool = False,
        **kwargs: Any,
    ) -> Generator[str, None, None]:
        raw = kwargs.get("raw", False)
        def stream_generator():
            gen = self.ask(
                prompt,
                stream=True,  # API always uses streaming
                raw=raw,
                optimizer=optimizer,
                conversationally=conversationally,
                **kwargs
            )
            for response in gen:
                if raw:
                    yield cast(str, response)
                else:
                    yield self.get_message(cast(Response, response))
        return stream_generator()

    def get_message(self, response: Response) -> str:
        if not isinstance(response, dict):
            return str(response)
        response_dict = cast(Dict[str, Any], response)
        return response_dict.get("text", "")


if __name__ == "__main__":
    from rich import print
    ai = TwoAI(api_key="", model="sutra-v2", temperature=0.7)
    response = ai.chat("Write a poem about AI in the style of Shakespeare.")
    for chunk in response:
        print(chunk, end="", flush=True)



import json
import time
import uuid
from typing import Any, Dict, Generator, List, Optional, Union, cast

import requests

from webscout.Provider.Openai_comp.base import (
    BaseChat,
    BaseCompletions,
    OpenAICompatibleProvider,
    SimpleModelList,
)
from webscout.Provider.Openai_comp.utils import (
    ChatCompletion,
    ChatCompletionChunk,
    ChatCompletionMessage,
    Choice,
    ChoiceDelta,
    CompletionUsage,
    count_tokens,
)

from ...litagent import LitAgent

# ANSI escape codes for formatting
BOLD = "\033[1m"
RED = "\033[91m"
RESET = "\033[0m"

# Model configurations
MODEL_CONFIGS = {
    "ayle": {
        "endpoint": "https://ayle.chat/api/chat",
        "models": [
            "gemini-2.5-flash",
            "llama-3.3-70b-versatile",
            "llama-3.3-70b",
            "tngtech/deepseek-r1t2-chimera:free",
            "openai/gpt-oss-120b",
            "qwen-3-235b-a22b-instruct-2507",
            "llama3.1-8b",
            "llama-4-scout-17b-16e-instruct",
            "qwen-3-32b",
        ],
    }
}


class Completions(BaseCompletions):
    def __init__(self, client: "Ayle"):
        self._client = client

    def create(
        self,
        *,
        model: str,
        messages: List[Dict[str, str]],
        max_tokens: Optional[int] = None,  # Not used directly but kept for compatibility
        stream: bool = False,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        timeout: Optional[int] = None,
        proxies: Optional[Dict[str, str]] = None,
        **kwargs: Any,
    ) -> Union[ChatCompletion, Generator[ChatCompletionChunk, None, None]]:
        """
        Creates a model response for the given chat conversation.
        Mimics openai.chat.completions.create
        """
        # Determine the provider based on the model
        provider = self._client._get_provider_from_model(model)

        # Build the appropriate payload
        payload = {"messages": messages, "model": model}

        request_id = f"chatcmpl-{uuid.uuid4()}"
        created_time = int(time.time())

        if stream:
            return self._create_stream(
                request_id, created_time, model, provider, payload, timeout, proxies
            )
        else:
            return self._create_non_stream(
                request_id, created_time, model, provider, payload, timeout, proxies
            )

    def _create_stream(
        self,
        request_id: str,
        created_time: int,
        model: str,
        provider: str,
        payload: Dict[str, Any],
        timeout: Optional[int] = None,
        proxies: Optional[Dict[str, str]] = None,
    ) -> Generator[ChatCompletionChunk, None, None]:
        try:
            endpoint = self._client._get_endpoint(provider)
            response = self._client.session.post(
                endpoint,
                headers=self._client.headers,
                json=payload,
                stream=True,
                timeout=timeout or self._client.timeout,
                proxies=proxies or getattr(self._client, "proxies", None),
            )
            response.raise_for_status()

            # Track token usage across chunks
            completion_tokens = 0
            streaming_text = ""

            for line in response.iter_lines():
                if not line:
                    continue

                try:
                    line_str = line.decode("utf-8")
                    if line_str.startswith('0:"'):
                        content = json.loads(line_str[2:])
                        if content:
                            streaming_text += content
                            completion_tokens += count_tokens(content)

                        # Create a delta object for this chunk
                        delta = ChoiceDelta(content=content)
                        choice = Choice(index=0, delta=delta, finish_reason=None)

                        chunk = ChatCompletionChunk(
                            id=request_id,
                            choices=[choice],
                            created=created_time,
                            model=model,
                        )

                        yield chunk
                except (json.JSONDecodeError, UnicodeDecodeError):
                    continue

            # Final chunk with finish_reason
            delta = ChoiceDelta(content=None)
            choice = Choice(index=0, delta=delta, finish_reason="stop")

            chunk = ChatCompletionChunk(
                id=request_id,
                choices=[choice],
                created=created_time,
                model=model,
            )

            yield chunk

        except requests.exceptions.RequestException as e:
            print(f"{RED}Error during Ayle stream request: {e}{RESET}")
            raise IOError(f"Ayle request failed: {e}") from e

    def _create_non_stream(
        self,
        request_id: str,
        created_time: int,
        model: str,
        provider: str,
        payload: Dict[str, Any],
        timeout: Optional[int] = None,
        proxies: Optional[Dict[str, str]] = None,
    ) -> ChatCompletion:
        try:
            endpoint = self._client._get_endpoint(provider)
            response = self._client.session.post(
                endpoint,
                headers=self._client.headers,
                json=payload,
                timeout=timeout or self._client.timeout,
                proxies=proxies or getattr(self._client, "proxies", None),
            )
            response.raise_for_status()

            full_response = ""
            for line in response.iter_lines():
                if line:
                    try:
                        line_str = line.decode("utf-8")
                        if line_str.startswith('0:"'):
                            content = json.loads(line_str[2:])
                            if content:
                                full_response += content
                    except (json.JSONDecodeError, UnicodeDecodeError):
                        continue

            # Create usage statistics (estimated)
            prompt_tokens = count_tokens(str(payload.get("messages", "")))
            completion_tokens = count_tokens(full_response)
            total_tokens = prompt_tokens + completion_tokens

            usage = CompletionUsage(
                prompt_tokens=prompt_tokens,
                completion_tokens=completion_tokens,
                total_tokens=total_tokens,
            )

            # Create the message object
            message = ChatCompletionMessage(role="assistant", content=full_response)

            # Create the choice object
            choice = Choice(index=0, message=message, finish_reason="stop")

            # Create the completion object
            completion = ChatCompletion(
                id=request_id,
                choices=[choice],
                created=created_time,
                model=model,
                usage=usage,
            )

            return completion

        except Exception as e:
            print(f"{RED}Error during Ayle non-stream request: {e}{RESET}")
            raise IOError(f"Ayle request failed: {e}") from e


class Chat(BaseChat):
    def __init__(self, client: "Ayle"):
        self.completions = Completions(client)


class Ayle(OpenAICompatibleProvider):
    """
    OpenAI-compatible client for Ayle API.

    Usage:
        client = Ayle()
        response = client.chat.completions.create(
            model="gemini-2.5-flash",
            messages=[{"role": "user", "content": "Hello!"}]
        )
        print(response.choices[0].message.content)
    """

    required_auth = False
    AVAILABLE_MODELS = [
        "gemini-2.5-flash",
        "llama-3.3-70b-versatile",
        "llama-3.3-70b",
        "tngtech/deepseek-r1t2-chimera:free",
        "openai/gpt-oss-120b",
        "qwen-3-235b-a22b-instruct-2507",
        "llama3.1-8b",
        "llama-4-scout-17b-16e-instruct",
        "qwen-3-32b",
    ]

    def __init__(self, timeout: int = 30, temperature: float = 0.5, top_p: float = 1.0):
        """
        Initialize the Ayle client.

        Args:
            timeout: Request timeout in seconds.
            temperature: Temperature for response generation.
            top_p: Top-p sampling parameter.
        """
        self.timeout = timeout
        self.temperature = temperature
        self.top_p = top_p

        # Initialize LitAgent for user agent generation
        agent = LitAgent()

        self.headers = {
            "accept": "*/*",
            "accept-language": "en-US,en;q=0.9",
            "content-type": "application/json",
            "origin": "https://ayle.chat/",
            "referer": "https://ayle.chat/",
            "user-agent": agent.random(),
        }

        self.session = requests.Session()
        self.session.headers.update(self.headers)
        self.session.cookies.update({"session": uuid.uuid4().hex})

        # Initialize the chat interface
        self.chat = Chat(self)

    @property
    def models(self) -> SimpleModelList:
        return SimpleModelList(type(self).AVAILABLE_MODELS)

    def _get_endpoint(self, provider: str) -> str:
        """Get the API endpoint for the specified provider."""
        return cast(str, MODEL_CONFIGS[provider]["endpoint"])

    def _get_provider_from_model(self, model: str) -> str:
        """Determine the provider based on the model name."""
        for provider, config in MODEL_CONFIGS.items():
            if model in config["models"]:
                return provider

        # If model not found, use a default model
        print(f"{BOLD}Warning: Model '{model}' not found, using default model 'ayle'{RESET}")
        return "ayle"

    def convert_model_name(self, model: str) -> str:
        """
        Ensure the model name is in the correct format.
        """
        if model in self.AVAILABLE_MODELS:
            return model

        # Try to find a matching model
        for available_model in self.AVAILABLE_MODELS:
            if model.lower() in available_model.lower():
                return available_model

        # Default to gemini-2.5-flash if no match
        print(
            f"{BOLD}Warning: Model '{model}' not found, using default model 'gemini-2.5-flash'{RESET}"
        )
        return "gemini-2.5-flash"


# Simple test if run directly
if __name__ == "__main__":
    print("-" * 80)
    print(f"{'Model':<50} {'Status':<10} {'Response'}")
    print("-" * 80)

    # Test a subset of models to avoid excessive API calls
    test_models = [
        "gemini-2.5-flash",
    ]

    for model in test_models:
        try:
            client = Ayle(timeout=60)
            # Test with a simple conversation to demonstrate format_prompt usage
            response = client.chat.completions.create(
                model=model,
                messages=[
                    {"role": "system", "content": "You are a helpful assistant."},
                    {"role": "user", "content": "Say 'Hello' in one word"},
                ],
                stream=False,
            )

            if (
                isinstance(response, ChatCompletion)
                and response.choices
                and response.choices[0].message
                and response.choices[0].message.content
            ):
                status = "✓"
                # Truncate response if too long
                display_text = response.choices[0].message.content.strip()
                display_text = display_text[:50] + "..." if len(display_text) > 50 else display_text
            else:
                status = "✗"
                display_text = "Empty or invalid response"
            print(f"{model:<50} {status:<10} {display_text}")
        except Exception as e:
            print(f"{model:<50} {'✗':<10} {str(e)}")
            print(f"{model:<50} {'✗':<10} {str(e)}")



# The `AwesomePrompts` class is a prompts manager in Python that fetches, caches, updates, and
# provides prompts with optimization features like LRU caching and concurrency.
# -*- coding: utf-8 -*-


import json
import threading
from datetime import datetime, timedelta
from functools import lru_cache
from pathlib import Path
from typing import Dict, Optional, Union

from curl_cffi.requests import Session
from rich.console import Console
from rich.table import Table

console = Console()


class AwesomePrompts:
    """Prompts manager with caching and optimization."""

    def __init__(
        self,
        repo_url: str = "https://raw.githubusercontent.com/OEvortex/prompts/main/prompt.json",
        local_path: Optional[str] = None,
        auto_update: bool = True,
        timeout: int = 10,
        impersonate: Optional[str] = "chrome110",
        cache_size: int = 128,
        max_workers: int = 4,
    ):
        """Initialize optimized Awesome Prompts.

        Args:
            repo_url: URL to fetch prompts from
            local_path: Where to save prompts locally
            auto_update: Auto update prompts on init
            timeout: Timeout for HTTP requests
            impersonate: Browser profile for curl_cffi
            cache_size: LRU cache size for get operations
            max_workers: Max threads for concurrent operations
        """
        self.repo_url = repo_url
        self.local_path = (
            Path(local_path) if local_path else Path.home() / ".webscout" / "awesome-prompts.json"
        )
        self.timeout = timeout
        self._last_update: Optional[datetime] = None
        self._cache_lock = threading.RLock()
        self._file_lock = threading.Lock()
        self._max_workers = max_workers

        self._max_workers = max_workers
        self.timeout = timeout
        try:
            self.session = Session(timeout=timeout, impersonate=impersonate)
        except Exception:
            self.session = Session(timeout=timeout)
        self.local_path.parent.mkdir(parents=True, exist_ok=True)
        self._cache: Dict[Union[str, int], str] = {}
        self._load_cache()
        self._get_cached = lru_cache(maxsize=cache_size)(self._get_uncached)
        if auto_update:
            self.update_prompts_from_online()

    def _load_cache(self) -> None:
        """Initialize cache from local file."""
        try:
            if self.local_path.exists():
                with self._file_lock:
                    with open(self.local_path, "r", encoding="utf-8") as f:
                        data = json.load(f)
                        with self._cache_lock:
                            self._cache = data
                        self._rebuild_numeric_indices()
        except (json.JSONDecodeError, IOError) as e:
            console.print(f"[red]Warning: Failed to load cache: {e}[/red]")
            self._cache = {}

    def _save_cache(self) -> None:
        """Save cache to local file with atomic write."""
        try:
            temp_path = self.local_path.with_suffix(".tmp")
            with self._file_lock:
                with open(temp_path, "w", encoding="utf-8") as f:
                    cache_to_save = {k: v for k, v in self._cache.items() if isinstance(k, str)}
                    json.dump(cache_to_save, f, indent=2, ensure_ascii=False, sort_keys=True)
                temp_path.replace(self.local_path)
        except IOError as e:
            console.print(f"[red]Error saving cache: {e}[/red]")

    def _load_prompts(self) -> Dict[Union[str, int], str]:
        """Load prompts from cache or file."""
        with self._cache_lock:
            if self._cache:
                return self._cache.copy()

        self._load_cache()
        self._rebuild_numeric_indices()
        with self._cache_lock:
            return self._cache.copy()

    def _rebuild_numeric_indices(self) -> None:
        """Rebuild numeric indices from string keys."""
        with self._cache_lock:
            numeric_keys = [k for k in self._cache.keys() if isinstance(k, int)]
            for key in numeric_keys:
                del self._cache[key]

            string_keys = [k for k in self._cache.keys() if isinstance(k, str)]
            for i, key in enumerate(string_keys):
                self._cache[i] = self._cache[key]

    def _save_prompts(self, prompts: Dict[Union[str, int], str]) -> None:
        """Save prompts and update cache."""
        with self._cache_lock:
            self._cache = prompts.copy()
        self._save_cache()

    def update_prompts_from_online(self, force: bool = False) -> bool:
        """Update prompts from repository with optimized merging."""
        try:
            if (
                not force
                and self._last_update
                and (datetime.now() - self._last_update) < timedelta(hours=1)
            ):
                console.print("[yellow]Prompts are already up to date![/yellow]")
                return True

            console.print("[cyan]Updating prompts...[/cyan]")

            response = self.session.get(self.repo_url)
            response.raise_for_status()

            new_prompts = response.json()
            if not isinstance(new_prompts, dict):
                raise ValueError("Invalid response format")

            existing_prompts = self._load_prompts()

            merged_prompts = {}
            string_keys = []

            for key, value in existing_prompts.items():
                if isinstance(key, str):
                    merged_prompts[key] = value
                    string_keys.append(key)

            for key, value in new_prompts.items():
                if isinstance(key, str):
                    merged_prompts[key] = value
                    if key not in string_keys:
                        string_keys.append(key)

            for i, key in enumerate(string_keys):
                merged_prompts[i] = merged_prompts[key]

            self._save_prompts(merged_prompts)
            self._last_update = datetime.now()

            console.print(
                f"[green]Updated {len([k for k in merged_prompts if isinstance(k, str)])} prompts successfully![/green]"
            )
            return True

        except Exception as e:
            error_msg = str(e)
            if hasattr(e, "response") and e.response is not None:
                status_code = getattr(e.response, "status_code", None)
                if status_code:
                    error_msg = f"HTTP {status_code}: {error_msg}"
            console.print(f"[red]Update failed: {error_msg}[/red]")
            return False

    def get_act(
        self,
        key: Union[str, int],
        default: Optional[str] = None,
        case_insensitive: bool = True,
        use_cache: bool = True,
    ) -> Optional[str]:
        """Get prompt with LRU caching for performance.

        Args:
            key: Prompt name or index
            default: Default value if not found
            case_insensitive: Enable case-insensitive matching
            use_cache: Use LRU cache for better performance
        """
        if use_cache:
            return self._get_cached(key, default, case_insensitive)
        return self._get_uncached(key, default, case_insensitive)

    def _get_uncached(
        self, key: Union[str, int], default: Optional[str] = None, case_insensitive: bool = True
    ) -> Optional[str]:
        """Core get logic without caching."""
        with self._cache_lock:
            prompts = self._cache if self._cache else self._load_prompts()

            if key in prompts:
                return prompts[key]

            if isinstance(key, str) and case_insensitive:
                key_lower = key.lower()
                for k, v in prompts.items():
                    if isinstance(k, str) and k.lower() == key_lower:
                        return v

        return default

    def add_prompt(self, name: str, prompt: str, validate: bool = True) -> bool:
        """Add a new prompt with validation and deduplication.

        Args:
            name: Name of the prompt
            prompt: The prompt text
            validate: Validate input and check for duplicates
        """
        if validate:
            if not name or not prompt:
                console.print("[red]Name and prompt cannot be empty![/red]")
                return False

            if len(name) > 100 or len(prompt) > 10000:
                console.print("[red]Name too long (max 100) or prompt too long (max 10000)[/red]")
                return False

        with self._cache_lock:
            prompts = self._load_prompts()

            if validate:
                for existing_name, existing_prompt in prompts.items():
                    if isinstance(existing_name, str) and existing_prompt == prompt:
                        console.print(
                            f"[yellow]Prompt with same content exists: '{existing_name}'[/yellow]"
                        )
                        return False

            prompts[name] = prompt

            string_keys = [k for k in prompts.keys() if isinstance(k, str)]
            for i, key in enumerate(string_keys):
                prompts[i] = prompts[key]

            self._save_prompts(prompts)

        console.print(f"[green]Added prompt: '{name}'[/green]")
        return True

    def delete_prompt(
        self, name: Union[str, int], case_insensitive: bool = True, raise_not_found: bool = False
    ) -> bool:
        """Delete a prompt with proper cleanup.

        Args:
            name: Name or index of prompt to delete
            case_insensitive: Enable case-insensitive matching
            raise_not_found: Raise error if prompt not found?
        """
        with self._cache_lock:
            prompts = self._load_prompts()

            if name in prompts:
                del prompts[name]

                string_keys = [k for k in prompts.keys() if isinstance(k, str)]
                numeric_keys = [k for k in prompts.keys() if isinstance(k, int)]
                for key in numeric_keys:
                    del prompts[key]

                for i, key in enumerate(string_keys):
                    prompts[i] = prompts[key]

                self._save_prompts(prompts)
                console.print(f"[green]Deleted prompt: '{name}'[/green]")
                return True

            if isinstance(name, str) and case_insensitive:
                name_lower = name.lower()
                for k in list(prompts.keys()):
                    if isinstance(k, str) and k.lower() == name_lower:
                        return self.delete_prompt(
                            k, case_insensitive=False, raise_not_found=raise_not_found
                        )

            if raise_not_found:
                raise KeyError(f"Prompt '{name}' not found!")
            console.print(f"[yellow]Prompt '{name}' not found![/yellow]")
            return False

    @property
    def all_acts(self) -> Dict[Union[str, int], str]:
        """Get all prompts with optimized indexing."""
        with self._cache_lock:
            if self._cache:
                return self._cache.copy()

        prompts = self._load_prompts()
        if not prompts:
            self.update_prompts_from_online()
            prompts = self._load_prompts()

        return prompts.copy()

    def show_acts(self, search: Optional[str] = None, limit: int = 100) -> None:
        """Display prompts with optimized filtering and pagination.

        Args:
            search: Filter by search term
            limit: Maximum number of prompts to display
        """
        prompts = self.all_acts

        filtered_items = []
        search_lower = search.lower() if search else None

        for key, value in prompts.items():
            if isinstance(key, int):
                continue

            if search_lower:
                if search_lower not in key.lower() and search_lower not in value.lower():
                    continue

            preview = value[:80] + "..." if len(value) > 80 else value
            filtered_items.append((str(key), preview))

            if len(filtered_items) >= limit:
                break

        if not filtered_items:
            console.print("[yellow]No prompts found[/yellow]")
            return

        table = Table(
            title=f"Awesome Prompts ({len(filtered_items)} shown)",
            show_header=True,
            header_style="bold magenta",
        )
        table.add_column("Name", style="green", max_width=30)
        table.add_column("Preview", style="yellow", max_width=50)

        for name, preview in filtered_items:
            table.add_row(name, preview)

        console.print(table)

    def get_random_act(self) -> Optional[str]:
        """Get a random prompt."""
        prompts = self.all_acts
        string_keys = [k for k in prompts.keys() if isinstance(k, str)]
        if not string_keys:
            return None
        import random

        return prompts[random.choice(string_keys)]


if __name__ == "__main__":
    prompt_manager = AwesomePrompts()
    print(prompt_manager.get_random_act())



import pathlib
import tempfile
import time
import urllib.parse
from concurrent.futures import ThreadPoolExecutor, as_completed
from io import BytesIO
from typing import Any, Optional, Union, cast

import requests
from litprinter import ic

from webscout import exceptions
from webscout.litagent import LitAgent

from . import utils
from .base import BaseTTSProvider


class StreamElements(BaseTTSProvider):

    """

    Text-to-speech provider using the StreamElements API.

    """

    required_auth = False



    # Supported voices

    SUPPORTED_VOICES = [

        "Filiz", "Astrid", "Tatyana", "Maxim", "Carmen", "Ines", "Cristiano",

        "Vitoria", "Ricardo", "Maja", "Jan", "Jacek", "Ewa", "Ruben", "Lotte",

        "Liv", "Seoyeon", "Takumi", "Mizuki", "Giorgio", "Carla", "Bianca",

        "Karl", "Dora", "Mathieu", "Celine", "Chantal", "Penelope", "Miguel",

        "Mia", "Enrique", "Conchita", "Geraint", "Salli", "Matthew", "Kimberly",

        "Kendra", "Justin", "Joey", "Joanna", "Ivy", "Raveena", "Aditi", "Emma",

        "Brian", "Amy", "Russell", "Nicole", "Vicki", "Marlene", "Hans", "Naja",

        "Mads", "Gwyneth", "Zhiyu", "es-ES-Standard-A", "it-IT-Standard-A",

        "it-IT-Wavenet-A", "ja-JP-Standard-A", "ja-JP-Wavenet-A", "ko-KR-Standard-A",

        "ko-KR-Wavenet-A", "pt-BR-Standard-A", "tr-TR-Standard-A", "sv-SE-Standard-A",

        "nl-NL-Standard-A", "nl-NL-Wavenet-A", "en-US-Wavenet-A", "en-US-Wavenet-B",

        "en-US-Wavenet-C", "en-US-Wavenet-D", "en-US-Wavenet-E", "en-US-Wavenet-F",

        "en-GB-Standard-A", "en-GB-Standard-B", "en-GB-Standard-C", "en-GB-Standard-D",

        "en-GB-Wavenet-A", "en-GB-Wavenet-B", "en-GB-Wavenet-C", "en-GB-Wavenet-D",

        "en-US-Standard-B", "en-US-Standard-C", "en-US-Standard-D", "en-US-Standard-E",

        "de-DE-Standard-A", "de-DE-Standard-B", "de-DE-Wavenet-A", "de-DE-Wavenet-B",

        "de-DE-Wavenet-C", "de-DE-Wavenet-D", "en-AU-Standard-A", "en-AU-Standard-B",

        "en-AU-Wavenet-A", "en-AU-Wavenet-B", "en-AU-Wavenet-C", "en-AU-Wavenet-D",

        "en-AU-Standard-C", "en-AU-Standard-D", "fr-CA-Standard-A", "fr-CA-Standard-B",

        "fr-CA-Standard-C", "fr-CA-Standard-D", "fr-FR-Standard-C", "fr-FR-Standard-D",

        "fr-FR-Wavenet-A", "fr-FR-Wavenet-B", "fr-FR-Wavenet-C", "fr-FR-Wavenet-D",

        "da-DK-Wavenet-A", "pl-PL-Wavenet-A", "pl-PL-Wavenet-B", "pl-PL-Wavenet-C",

        "pl-PL-Wavenet-D", "pt-PT-Wavenet-A", "pt-PT-Wavenet-B", "pt-PT-Wavenet-C",

        "pt-PT-Wavenet-D", "ru-RU-Wavenet-A", "ru-RU-Wavenet-B", "ru-RU-Wavenet-C",

        "ru-RU-Wavenet-D", "sk-SK-Wavenet-A", "tr-TR-Wavenet-A", "tr-TR-Wavenet-B",

        "tr-TR-Wavenet-C", "tr-TR-Wavenet-D", "tr-TR-Wavenet-E", "uk-UA-Wavenet-A",

        "ar-XA-Wavenet-A", "ar-XA-Wavenet-B", "ar-XA-Wavenet-C", "cs-CZ-Wavenet-A",

        "nl-NL-Wavenet-B", "nl-NL-Wavenet-C", "nl-NL-Wavenet-D", "nl-NL-Wavenet-E",

        "en-IN-Wavenet-A", "en-IN-Wavenet-B", "en-IN-Wavenet-C", "fil-PH-Wavenet-A",

        "fi-FI-Wavenet-A", "el-GR-Wavenet-A", "hi-IN-Wavenet-A", "hi-IN-Wavenet-B",

        "hi-IN-Wavenet-C", "hu-HU-Wavenet-A", "id-ID-Wavenet-A", "id-ID-Wavenet-B",

        "id-ID-Wavenet-C", "it-IT-Wavenet-B", "it-IT-Wavenet-C", "it-IT-Wavenet-D",

        "ja-JP-Wavenet-B", "ja-JP-Wavenet-C", "ja-JP-Wavenet-D", "cmn-CN-Wavenet-A",

        "cmn-CN-Wavenet-B", "cmn-CN-Wavenet-C", "cmn-CN-Wavenet-D", "nb-no-Wavenet-E",

        "nb-no-Wavenet-A", "nb-no-Wavenet-B", "nb-no-Wavenet-C", "nb-no-Wavenet-D",

        "vi-VN-Wavenet-A", "vi-VN-Wavenet-B", "vi-VN-Wavenet-C", "vi-VN-Wavenet-D",

        "sr-rs-Standard-A", "lv-lv-Standard-A", "is-is-Standard-A", "bg-bg-Standard-A",

        "af-ZA-Standard-A", "Tracy", "Danny", "Huihui", "Yaoyao", "Kangkang", "HanHan",

        "Zhiwei", "Asaf", "An", "Stefanos", "Filip", "Ivan", "Heidi", "Herena",

        "Kalpana", "Hemant", "Matej", "Andika", "Rizwan", "Lado", "Valluvar",

        "Linda", "Heather", "Sean", "Michael", "Karsten", "Guillaume", "Pattara",

        "Jakub", "Szabolcs", "Hoda", "Naayf"

    ]

    all_voices = SUPPORTED_VOICES

    # Request headers

    headers: dict[str, str] = {

        "User-Agent": LitAgent().random()

    }



    def __init__(self, timeout: int = 20, proxies: Optional[dict] = None):
        """Initializes the StreamElements TTS client."""
        super().__init__()
        self.session = requests.Session()
        self.session.headers.update(self.headers)
        if proxies:
            self.session.proxies.update(proxies)
        self.timeout = timeout

    def tts(self, text: str, voice: Optional[str] = None, verbose: bool = False, **kwargs) -> str:
        """
        Converts text to speech using the StreamElements API and saves it to a file.

        Args:
            text (str): The text to convert to speech
            voice (str): The voice to use for TTS
            verbose (bool): Whether to print progress messages
            **kwargs: Additional parameters

        Returns:
            str: Path to the generated audio file
        """
        voice = voice or kwargs.get('voice', "Emma")
        verbose = verbose if verbose is not None else kwargs.get('verbose', True)
        if voice not in self.all_voices:
             # Try case-insensitive match
             found_voice = None
             for v in self.all_voices:
                 if v.lower() == voice.lower():
                     found_voice = v
                     break
             if found_voice:
                 voice = found_voice
             else:
                 raise ValueError(f"Voice '{voice}' not one of [{', '.join(self.all_voices)}]")

        filename = pathlib.Path(tempfile.NamedTemporaryFile(suffix=".mp3", dir=self.temp_dir, delete=False).name)

        # Split text into sentences
        sentences = utils.split_sentences(text)

        # Function to request audio for each chunk
        def generate_audio_for_chunk(part_text: str, part_number: int):
            while True:
                try:
                    # URL encode the text and voice
                    encoded_text = urllib.parse.quote(part_text)
                    encoded_voice = urllib.parse.quote(voice)

                    url = f"https://streamelements.com/tts/{encoded_voice}/{encoded_text}"

                    response = self.session.get(url, headers=self.headers, timeout=self.timeout)
                    response.raise_for_status()

                    # Check if the request was successful
                    if response.ok and response.status_code == 200:
                        if verbose:
                            ic.configureOutput(prefix='DEBUG| ')
                            ic(f"Chunk {part_number} processed successfully")
                        return part_number, response.content
                    else:
                        if verbose:
                            ic.configureOutput(prefix='DEBUG| ')
                            ic(f"No data received for chunk {part_number}. Retrying...")
                except requests.RequestException as e:
                    if verbose:
                        ic.configureOutput(prefix='DEBUG| ')
                        ic(f"Error for chunk {part_number}: {e}. Retrying...")
                    time.sleep(1)
        try:
            # Using ThreadPoolExecutor to handle requests concurrently
            with ThreadPoolExecutor() as executor:
                futures = {executor.submit(generate_audio_for_chunk, sentence.strip(), chunk_num): chunk_num
                        for chunk_num, sentence in enumerate(sentences, start=1)}

                # Dictionary to store results with order preserved
                audio_chunks = {}

                for future in as_completed(futures):
                    chunk_num = futures[future]
                    try:
                        part_number, audio_data = future.result()
                        audio_chunks[part_number] = audio_data  # Store the audio data in correct sequence
                    except Exception as e:
                        if verbose:
                            ic.configureOutput(prefix='DEBUG| ')
                            ic(f"Failed to generate audio for chunk {chunk_num}: {e}")

            # Combine audio chunks in the correct sequence
            combined_audio = BytesIO()
            for part_number in sorted(audio_chunks.keys()):
                combined_audio.write(audio_chunks[part_number])
                if verbose:
                    ic.configureOutput(prefix='DEBUG| ')
                    ic(f"Added chunk {part_number} to the combined file.")

            # Save the combined audio data to a single file
            with open(filename, 'wb') as f:
                f.write(combined_audio.getvalue())
            if verbose:
                ic.configureOutput(prefix='DEBUG| ')
                ic(f"Final Audio Saved as {filename}")
            return filename.as_posix()

        except requests.exceptions.RequestException as e:
            if verbose:
                ic.configureOutput(prefix='DEBUG| ')
                ic(f"Failed to perform the operation: {e}")
            raise exceptions.FailedToGenerateResponseError(
                f"Failed to perform the operation: {e}"
            )

    def create_speech(
        self,
        input_text: str,
        model: Optional[str] = "gpt-4o-mini-tts",
        voice: Optional[str] = "Emma",
        response_format: Optional[str] = "mp3",
        instructions: Optional[str] = None,
        verbose: bool = False
    ) -> str:
        """
        OpenAI-compatible speech creation interface.
        """
        return self.tts(
            text=input_text,
            model=model,
            voice=voice or "Emma",
            response_format=response_format or "mp3",
            instructions=instructions,
            verbose=verbose
        )

# Example usage
if __name__ == "__main__":
    streamelements = StreamElements()
    text = "This is a test of the StreamElements text-to-speech API. It supports multiple sentences and advanced logging."

    ic.configureOutput(prefix='DEBUG| ')
    ic("Generating audio...")
    audio_file = streamelements.tts(text, voice="Mathieu")
    ic.configureOutput(prefix='INFO| ')
    ic(f"Audio saved to: {audio_file}")



"""DuckDuckGo image search."""

from __future__ import annotations

from ....search.results import ImagesResult
from .base import DuckDuckGoBase


class DuckDuckGoImages(DuckDuckGoBase):
    """DuckDuckGo image search."""

    name = "duckduckgo"
    category = "images"

    def run(self, *args, **kwargs) -> list[ImagesResult]:
        """Perform image search on DuckDuckGo.

        Args:
            keywords: Search query.
            region: Region code.
            safesearch: on, moderate, or off.
            timelimit: d, w, m, or y.
            size: Small, Medium, Large, Wallpaper.
            color: color name or Monochrome.
            type_image: photo, clipart, gif, transparent, line.
            layout: Square, Tall, Wide.
            license_image: any, Public, Share, etc.
            max_results: Maximum number of results.

        Returns:
            List of ImagesResult objects.
        """
        keywords = args[0] if args else kwargs.get("keywords")
        region = args[1] if len(args) > 1 else kwargs.get("region", "wt-wt")
        safesearch = args[2] if len(args) > 2 else kwargs.get("safesearch", "moderate")
        timelimit = args[3] if len(args) > 3 else kwargs.get("timelimit")
        size = args[4] if len(args) > 4 else kwargs.get("size")
        color = args[5] if len(args) > 5 else kwargs.get("color")
        type_image = args[6] if len(args) > 6 else kwargs.get("type_image")
        layout = args[7] if len(args) > 7 else kwargs.get("layout")
        license_image = args[8] if len(args) > 8 else kwargs.get("license_image")
        max_results = args[9] if len(args) > 9 else kwargs.get("max_results")

        assert keywords, "keywords is mandatory"

        vqd = self._get_vqd(keywords)

        safesearch_base = {"on": "1", "moderate": "1", "off": "-1"}
        timelimit = f"time:{timelimit}" if timelimit else ""
        size = f"size:{size}" if size else ""
        color = f"color:{color}" if color else ""
        type_image = f"type:{type_image}" if type_image else ""
        layout = f"layout:{layout}" if layout else ""
        license_image = f"license:{license_image}" if license_image else ""
        payload = {
            "l": region,
            "o": "json",
            "q": keywords,
            "vqd": vqd,
            "f": f"{timelimit},{size},{color},{type_image},{layout},{license_image}",
            "p": safesearch_base[safesearch.lower()],
        }

        cache = set()
        results: list[ImagesResult] = []

        def _images_page(s: int) -> list[ImagesResult]:
            payload["s"] = f"{s}"
            resp_content = self._get_url("GET", "https://duckduckgo.com/i.js", params=payload).content
            resp_json = self.json_loads(resp_content)

            page_data = resp_json.get("results", [])
            page_results = []
            for row in page_data:
                image_url = row.get("image")
                if image_url and image_url not in cache:
                    cache.add(image_url)
                    result = ImagesResult(
                        title=row["title"],
                        image=self._normalize_url(image_url),
                        thumbnail=self._normalize_url(row["thumbnail"]),
                        url=self._normalize_url(row["url"]),
                        height=row["height"],
                        width=row["width"],
                        source=row["source"],
                    )
                    page_results.append(result)
            return page_results

        slist = [0]
        if max_results:
            max_results = min(max_results, 500)
            slist.extend(range(100, max_results, 100))
        try:
            for r in self._executor.map(_images_page, slist):
                results.extend(r)
        except Exception as e:
            raise e

        return list(self.islice(results, max_results))



"""DuckDuckGo search engines package."""

from .answers import DuckDuckGoAnswers
from .base import DuckDuckGoBase
from .images import DuckDuckGoImages
from .maps import DuckDuckGoMaps
from .news import DuckDuckGoNews
from .suggestions import DuckDuckGoSuggestions
from .text import DuckDuckGoTextSearch
from .translate import DuckDuckGoTranslate
from .videos import DuckDuckGoVideos
from .weather import DuckDuckGoWeather

__all__ = [
    "DuckDuckGoBase",
    "DuckDuckGoTextSearch",
    "DuckDuckGoImages",
    "DuckDuckGoVideos",
    "DuckDuckGoNews",
    "DuckDuckGoAnswers",
    "DuckDuckGoSuggestions",
    "DuckDuckGoMaps",
    "DuckDuckGoTranslate",
    "DuckDuckGoWeather",
]



import base64
import json
import os
import random
from typing import Any, Dict, Generator, Optional, Union, cast

from curl_cffi import CurlError
from curl_cffi.const import CurlHttpVersion
from curl_cffi.requests import Session

from webscout import exceptions
from webscout.AIbase import Provider
from webscout.AIutel import AwesomePrompts, Conversation, Optimizers
from webscout.litagent import LitAgent


class GizAI(Provider):
    """
    A class to interact with the GizAI API.

    Attributes:
        system_prompt (str): The system prompt to define the assistant's role.

    Examples:
        >>> from webscout.Provider.GizAI import GizAI
        >>> ai = GizAI()
        >>> response = ai.chat("What's the weather today?")
        >>> print(response)
    """
    required_auth = False
    AVAILABLE_MODELS = [
        "azure-gpt-4-1",
        "chat-gpt4",
        "chat-grok-2",
        "chat-o4-mini",
        "chat-o4-mini-high",
        "chat-o4-mini-medium",
        "claude-haiku",
        "claude-sonnet",
        "deepinfra-llama-4-maverick",
        "deepseek",
        "deepseek-r1-distill-llama-70b",
        "gemini-2.0-flash-lite",
        "gemini-2.5-flash",
        "gemini-2.5-pro",
        "gpt-4-1-mini",
        "gpt-4-1-nano",
        "gpt-4o-image",
        "hyperbolic-deepseek-r1",
        "llama-3-70b",
        "llama-4-scout",
        "o3",
        "phi-4",
        "qwq-32b"
    ]

    def __init__(
        self,
        is_conversation: bool = True,
        max_tokens: int = 2049,
        timeout: int = 30,
        intro: Optional[str] = None,
        filepath: Optional[str] = None,
        update_file: bool = True,
        proxies: dict = {},
        history_offset: int = 10250,
        act: Optional[str] = None,
        model: str = "gemini-2.0-flash-lite",
        system_prompt: str = "You are a helpful assistant."
    ):
        """Initializes the GizAI API client."""
        if model not in self.AVAILABLE_MODELS:
            raise ValueError(f"Invalid model: {model}. Choose from: {self.AVAILABLE_MODELS}")

        self.api_url = "https://app.giz.ai/api/data/users/inferenceServer.infer"

        # Initialize LitAgent for user-agent generation
        self.agent = LitAgent()

        # Initialize curl_cffi Session
        self.session = Session()

        # Set up the headers
        self.headers = {
            "accept": "application/json, text/plain, */*",
            "content-type": "application/json",
            "user-agent": self.agent.random(),
            "origin": "https://app.giz.ai",
            "referer": "https://app.giz.ai/",
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin"
        }

        # Update session headers and proxies
        self.session.headers.update(self.headers)
        if proxies:
            self.session.proxies.update(proxies)

        # Store configuration
        self.system_prompt = system_prompt
        self.is_conversation = is_conversation
        self.max_tokens_to_sample = max_tokens
        self.timeout = timeout
        self.last_response = {}
        self.model = model

        self.__available_optimizers = (
            method
            for method in dir(Optimizers)
            if callable(getattr(Optimizers, method)) and not method.startswith("__")
        )

        self.conversation = Conversation(
            is_conversation, self.max_tokens_to_sample, filepath, update_file
        )
        act_prompt = (
            AwesomePrompts().get_act(cast(Union[str, int], act), default=None, case_insensitive=True
            )
            if act
            else intro
        )
        if act_prompt:
            self.conversation.intro = act_prompt
        self.conversation.history_offset = history_offset

    def _generate_id(self, length: int = 21) -> str:
        """Generates a random URL-safe base64 string."""
        random_bytes = os.urandom(length * 2)  # Generate more bytes initially
        b64_encoded = base64.urlsafe_b64encode(random_bytes).decode('utf-8')
        return b64_encoded[:length]

    def _get_random_ip(self) -> str:
        """Generates a random IPv4 address string."""
        return f"{random.randint(0, 255)}.{random.randint(0, 255)}.{random.randint(0, 255)}.{random.randint(0, 255)}"

    def ask(
        self,
        prompt: str,
        stream: bool = False,
        raw: bool = False,
        optimizer: Optional[str] = None,
        conversationally: bool = False,
        **kwargs: Any,
    ) -> Union[Dict[str, Any], str]:
        """
        Sends a prompt to the GizAI API and returns the response.

        Args:
            prompt (str): The prompt to send to the API.
            stream (bool): Not supported by GizAI, kept for compatibility.
            raw (bool): Whether to return the raw response.
            optimizer (str): Optimizer to use for the prompt.
            conversationally (bool): Whether to generate the prompt conversationally.

        Returns:
            Dict[str, Any]: The API response.

        Examples:
            >>> ai = GizAI()
            >>> response = ai.ask("Tell me a joke!")
        """
        conversation_prompt = self.conversation.gen_complete_prompt(prompt)
        if optimizer:
            if optimizer in self.__available_optimizers:
                conversation_prompt = getattr(Optimizers, optimizer)(
                    conversation_prompt if conversationally else prompt
                )
            else:
                raise Exception(f"Optimizer is not one of {self.__available_optimizers}")

        # Generate random IDs for request
        instance_id = self._generate_id()
        subscribe_id = self._generate_id()
        x_forwarded_for = self._get_random_ip()

        # Set up request body - GizAI doesn't support streaming
        request_body = {
            "model": "chat",
            "baseModel": self.model, # Use the specific model ID here
            "input": {
                "messages": [{
                    "type": "human",
                    "content": conversation_prompt
                }],
                "mode": "plan"
            },
            "noStream": True,
            "instanceId": instance_id,
            "subscribeId": subscribe_id
        }

        # Combine default headers with the dynamic x-forwarded-for header
        request_headers = {**self.headers, "x-forwarded-for": x_forwarded_for}

        try:
            # Use curl_cffi session post with impersonate
            response = self.session.post(
                self.api_url,
                headers=request_headers,
                json=request_body,
                timeout=self.timeout,
                impersonate="chrome120",  # Use a common impersonation profile
                http_version=CurlHttpVersion.V2_0    # Use HTTP/2
            )
            response.raise_for_status()  # Check for HTTP errors

            # Process the response
            try:
                response_json = response.json()
                # GizAI responses have "status" and "output" fields
                if response_json.get("status") == "completed" and "output" in response_json:
                    content = response_json["output"]
                else:
                    content = ""
                    # Try to extract content from any available field that might contain the response
                    for key, value in response_json.items():
                        if isinstance(value, str) and len(value) > 10:
                            content = value
                            break
            except json.JSONDecodeError:
                # Handle case where response is not valid JSON
                content = response.text

            # Update conversation history
            self.last_response = {"text": content}
            self.conversation.update_chat_history(prompt, content)

            return self.last_response if not raw else content

        except CurlError as e:
            raise exceptions.FailedToGenerateResponseError(f"Request failed (CurlError): {str(e)}")
        except Exception as e:
            error_text = getattr(e, 'response', None) and getattr(e.response, 'text', '') if hasattr(e, 'response') else ''
            raise exceptions.FailedToGenerateResponseError(f"Request failed ({type(e).__name__}): {str(e)} - {error_text}")

    def chat(
        self,
        prompt: str,
        stream: bool = False,
        optimizer: Optional[str] = None,
        conversationally: bool = False,
        **kwargs: Any,
    ) -> Union[str, Generator[str, None, None]]:
        """
        Generates a response from the GizAI API.

        Args:
            prompt (str): The prompt to send to the API.
            stream (bool): Not supported by GizAI, kept for compatibility.
            optimizer (str): Optimizer to use for the prompt.
            conversationally (bool): Whether to generate the prompt conversationally.

        Returns:
            Generator[str, None, None]: The API response text as a generator.

        Examples:
            >>> ai = GizAI()
            >>> response = ai.chat("What's the weather today?")
        """
        # GizAI doesn't support streaming, so ignore the stream parameter
        response_data = self.ask(
            prompt, stream=False, raw=False,
            optimizer=optimizer, conversationally=conversationally
        )
        result = self.get_message(response_data)
        if stream:
            yield result
        else:
            return result

    def get_message(self, response: Union[Dict[str, Any], Generator[Any, None, None], str]) -> str:
        """
        Extracts the message from the API response.

        Args:
            response (Union[dict, str]): The API response.

        Returns:
            str: The message content.

        Examples:
            >>> ai = GizAI()
            >>> response = ai.ask("Tell me a joke!")
            >>> message = ai.get_message(response)
        """
        if isinstance(response, str):
            return response
        elif isinstance(response, dict):
            resp_dict = cast(Dict[str, Any], response)
            return resp_dict.get("text", "")
        else:
            # Generator case - shouldn't happen
            return ""

if __name__ == "__main__":
    ai = GizAI()
    response = ai.chat("Hello, how are you?", stream=True)
    for chunk in response:
        print(chunk, end="", flush=True)



# webscout/__init__.py

from .AIauto import *  # noqa: F403
from .AIutel import *  # noqa: F403
from .client import Client
from .Extra import *  # noqa: F403
from .litagent import LitAgent
from .models import model
from .optimizers import *
from .Provider import *
from .Provider.AISEARCH import *
from .Provider.STT import *  # noqa: F403
from .Provider.TTI import *
from .Provider.TTS import *
from .scout import *
from .search import *
from .swiftcli import *
from .update_checker import check_for_updates
from .version import __version__
from .zeroart import *

useragent = LitAgent()
try:
    update_message = check_for_updates(force=False)
    if update_message:
        print(update_message)
except Exception:
    pass




from .gist import Gist
from .organization import Organization
from .repository import Repository
from .search import GitSearch
from .trending import Trending
from .user import User
from .utils import GitError, NotFoundError, RateLimitError, RequestError

__all__ = [
    'Repository',
    'User',
    'GitSearch',
    'Gist',
    'Organization',
    'Trending',
    'GitError',
    'RateLimitError',
    'NotFoundError',
    'RequestError'
]



import functools
import time

# --- Utility Decorators ---
from typing import Callable, Optional

from .conversation import Conversation  # noqa: E402,F401
from .optimizers import Optimizers  # noqa: E402,F401
from .prompt_manager import AwesomePrompts  # noqa: E402,F401
from .sanitize import *  # noqa: E402, F401, F403


def timeIt(func: Callable):
    """
    Decorator to measure execution time of a function (sync or async).
    Prints: - Execution time for '{func.__name__}' : {elapsed:.6f} Seconds.
    """
    import asyncio
    GREEN_BOLD = "\033[1;92m"
    RESET = "\033[0m"
    @functools.wraps(func)
    def sync_wrapper(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()
        print()
        func_name = getattr(func, "__name__", str(func))
        print(f"{GREEN_BOLD}- Execution time for '{func_name}' : {end_time - start_time:.6f} Seconds.  {RESET}\n")
        return result

    @functools.wraps(func)
    async def async_wrapper(*args, **kwargs):
        start_time = time.time()
        result = await func(*args, **kwargs)
        end_time = time.time()
        print()
        func_name = getattr(func, "__name__", str(func))
        print(f"{GREEN_BOLD}- Execution time for '{func_name}' : {end_time - start_time:.6f} Seconds.  {RESET}\n")
        return result

    if asyncio.iscoroutinefunction(func):
        return async_wrapper
    else:
        return sync_wrapper

def retry(retries: int = 3, delay: float = 1) -> Callable:
    """
    Decorator to retry a function on exception.
    """
    def decorator(func: Callable):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            last_exc = None
            for attempt in range(retries):
                try:
                    return func(*args, **kwargs)
                except Exception as exc:
                    last_exc = exc
                    print(f"Attempt {attempt + 1} failed: {exc}. Retrying in {delay} seconds...")
                    time.sleep(delay)
            # Ensure we have an exception to raise
            if last_exc is not None:
                raise last_exc
            else:
                func_name = getattr(func, "__name__", str(func))
                raise RuntimeError(f"Function {func_name} failed after {retries} retries with no exception recorded")
        return wrapper
    return decorator
